__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import expression, generic, global_constants, nmlop
from nml.actions import (
    action0,
    action2,
    action2layout,
    action2real,
    action2var,
    action3_callbacks,
    action6,
    actionD,
    base_action,
)


class Action3(base_action.BaseAction):
    """
    Class representing a single Action3.

    @ivar feature: Action3 feature byte.
    @type feature: C{int}

    @ivar id: Item ID of the item that this action3 represents.
    @type id: C{int}

    @ivar is_livery_override: Whether this action 3 is a livery override
    @type is_livery_override: C{bool}

    @ivar cid_mappings: List of mappings that map cargo IDs to Action2s.
    @type cid_mappings: C{list} of C{tuple}: (C{int}, L{Expression} before prepare_output, C{int} afterwards, C{str})

    @ivar def_cid: Default Action2 to use if no cargo ID matches.
    @type def_cid: C{None} or L{SpriteGroupRef} before prepare_output, C{int} afterwards

    @ivar references: All Action2s that are referenced by this Action3.
    @type references: C{list} of L{Action2Reference}
    """

    def __init__(self, feature, id, is_livery_override):
        self.feature = feature
        self.id = id
        self.is_livery_override = is_livery_override
        self.cid_mappings = []
        self.references = []

    def prepare_output(self, sprite_num):
        action2.free_references(self)

        def map_cid(cid):
            if isinstance(cid, expression.SpriteGroupRef):
                return cid.get_action2_id(self.feature)
            else:
                return cid.value | 0x8000

        self.cid_mappings = [(cargo, map_cid(cid), comment) for cargo, cid, comment in self.cid_mappings]
        if self.def_cid is None:
            self.def_cid = 0
        else:
            self.def_cid = map_cid(self.def_cid)

    def write(self, file):
        size = 7 + 3 * len(self.cid_mappings)
        if self.feature <= 3 or self.id >= 0xFF:
            size += 2  # Vehicles or IDs >= 255 use extended byte
        file.start_sprite(size)
        file.print_bytex(3)
        file.print_bytex(self.feature)
        file.print_bytex(1 if not self.is_livery_override else 0x81)  # a single id
        file.print_varx(self.id, 3 if self.feature <= 3 or self.id >= 0xFF else 1)
        file.print_byte(len(self.cid_mappings))
        file.newline()
        for cargo, cid, comment in self.cid_mappings:
            file.print_bytex(cargo)
            file.print_wordx(cid)
            file.newline(comment)
        file.print_wordx(self.def_cid)
        file.newline(self.default_comment)
        file.end_sprite()

    def skip_action9(self):
        return False


# Make sure all action2s created here get a unique name
action2_id = 0


def create_intermediate_varaction2(feature, varact2parser, mapping, default, pos):
    """
    Create a varaction2 based on a parsed expression and a value mapping

    @param feature: Feature of the varaction2
    @type feature: C{int}

    @param varact2parser: Parser containing a parsed expression
    @type varact2parser: L{Varaction2Parser}

    @param mapping: Mapping of various values to sprite groups / return values,
                    with a possible extra function to apply to the return value
    @type mapping: C{dict} that maps C{int} to C{tuple} of (L{SpriteGroupRef}, C{function}, or C{None})

    @param default: Default sprite group if no value matches
    @type default: L{SpriteGroupRef}

    @param pos: Positional context.
    @type  pos: L{Position}

    @return: A tuple containing the action list and a reference to the created action2
    @rtype: C{tuple} of (C{list} of L{BaseAction}, L{SpriteGroupRef})
    """
    global action2_id

    action_list = varact2parser.extra_actions
    act6 = action6.Action6()
    for mod in varact2parser.mods:
        act6.modify_bytes(mod.param, mod.size, mod.offset + 4)

    name = expression.Identifier("@action3_{:d}".format(action2_id))
    action2_id += 1
    varaction2 = action2var.Action2Var(feature, name.value, pos, 0x89)
    varaction2.var_list = varact2parser.var_list
    offset = 5 + varact2parser.var_list_size
    for proc in varact2parser.proc_call_list:
        action2.add_ref(proc, varaction2, True)

    for switch_value in sorted(mapping):
        return_value, ret_value_function = mapping[switch_value]
        if ret_value_function is None:
            result, comment = action2var.parse_result(return_value, action_list, act6, offset, varaction2, None, 0x89)
        else:
            if isinstance(return_value, expression.SpriteGroupRef):
                # We need to execute the callback via a procedure call
                # then return CB_FAILED if the CB failed,
                # or the CB result (with ret_value_function applied) if successful
                if return_value.name.value == "CB_FAILED":
                    result, comment = action2var.parse_result(
                        return_value, action_list, act6, offset, varaction2, None, 0x89
                    )
                else:
                    extra_actions, result, comment = create_proc_call_varaction2(
                        feature, return_value, ret_value_function, varaction2, pos
                    )
                    action_list.extend(extra_actions)
            else:
                return_value = ret_value_function(return_value).reduce()
                result, comment = action2var.parse_result(
                    return_value, action_list, act6, offset, varaction2, None, 0x89
                )

        varaction2.ranges.append(
            action2var.VarAction2Range(
                expression.ConstantNumeric(switch_value), expression.ConstantNumeric(switch_value), result, comment
            )
        )
        offset += 10
    result, comment = action2var.parse_result(default, action_list, act6, offset, varaction2, None, 0x89)
    varaction2.default_result = result
    varaction2.default_comment = comment

    return_ref = expression.SpriteGroupRef(name, [], None, varaction2)
    if len(act6.modifications) > 0:
        action_list.append(act6)
    action_list.append(varaction2)
    return (action_list, return_ref)


def create_proc_call_varaction2(feature, proc, ret_value_function, parent_action, pos):
    """
    Create a varaction2 that executes a procedure call and applies a function on the result

    @param feature: Feature of the varaction2
    @type feature: C{int}

    @param proc: Procedure to execute
    @type proc: L{SpriteGroupRef}

    @param ret_value_function: Function to apply on the result (L{Expression} -> L{Expression})
    @type ret_value_function: C{function}

    @param pos: Positional context.
    @type  pos: L{Position}

    @return: A list of extra actions, reference to the created action2 and a comment to add
    @rtype: C{tuple} of (C{list} of L{BaseAction}, L{SpriteGroupRef}, C{str})
    """
    proc.is_procedure = True
    varact2parser = action2var.Varaction2Parser(feature)
    varact2parser.parse_proc_call(proc)

    mapping = {0xFFFF: (expression.SpriteGroupRef(expression.Identifier("CB_FAILED"), [], None), None)}
    default = ret_value_function(expression.Variable(expression.ConstantNumeric(0x1C)))
    action_list, result = create_intermediate_varaction2(feature, varact2parser, mapping, default, pos)
    action2.add_ref(result, parent_action)
    comment = result.name.value + ";"
    return (action_list, result, comment)


def create_cb_choice_varaction2(feature, expr, mapping, default, pos):
    """
    Create a varaction2 that maps callback numbers to various sprite groups

    @param feature: Feature of the varaction2
    @type feature: C{int}

    @param expr: Expression to evaluate
    @type expr: L{Expression}

    @param mapping: Mapping of various values to sprite groups,
                    with a possible extra function to apply to the return value
    @type mapping: C{dict} that maps C{int} to C{tuple} of (L{SpriteGroupRef}, C{function}, or C{None})

    @param default: Default sprite group if no value matches
    @type default: L{SpriteGroupRef}

    @param pos: Positional context.
    @type  pos: L{Position}

    @return: A tuple containing the action list and a reference to the created action2
    @rtype: C{tuple} of (C{list} of L{BaseAction}, L{SpriteGroupRef})
    """
    varact2parser = action2var.Varaction2Parser(feature)
    varact2parser.parse_expr(expr)
    return create_intermediate_varaction2(feature, varact2parser, mapping, default, pos)


def create_action3(feature, id, action_list, act6, is_livery_override):
    # Vehicles use an extended byte
    size = 2 if feature <= 3 else 1
    offset = 4 if feature <= 3 else 3

    id, offset = actionD.write_action_value(id, action_list, act6, offset, size)
    return Action3(feature, id.value, is_livery_override)


house_tiles = {
    0: "n",  # 1x1
    2: "nw",  # 2x1
    3: "ne",  # 1x2
    4: "news",  # 2x2
}


def parse_graphics_block(graphics_block, feature, id, size, is_livery_override=False):
    """
    Parse a graphics block (or livery override) into a list of actions, mainly action3

    @param graphics_block: Graphics-block to parse
    @type graphics_block: L{GraphicsBlock}

    @param feature: Feature of the associated item
    @type feature: C{int}

    @param id: ID of the associated item
    @type id: L{Expression}

    @param size: Size of the associated item (relevant for houses only)
    @type size: L{ConstantNumeric} or C{None}

    @param is_livery_override: Whether this is a livery override instead of a normal graphics block
    @type is_livery_override: C{bool}

    @return: The resulting list of actions
    @rtype: L{BaseAction}
    """
    action_list = action2real.create_spriteset_actions(graphics_block)
    if feature == 0x07:
        # Multi-tile houses need more work
        size_bit = size.value if size is not None else 0
        for i, tile in enumerate(house_tiles[size_bit]):
            tile_id = id if i == 0 else nmlop.ADD(id, i).reduce()
            action_list.extend(
                parse_graphics_block_single_id(graphics_block, feature, tile_id, is_livery_override, tile, id)
            )
    else:
        action_list.extend(parse_graphics_block_single_id(graphics_block, feature, id, is_livery_override))
    return action_list


station_sprite_layouts = {}


def parse_graphics_block_single_id(
    graphics_block, feature, id, is_livery_override, house_tile=None, house_north_tile_id=None
):
    action6.free_parameters.save()
    prepend_action_list = []
    action_list = []
    act6 = action6.Action6()
    act3 = create_action3(feature, id, action_list, act6, is_livery_override)

    cargo_gfx = {}
    seen_callbacks = set()
    callbacks = []
    livery_override = None  # Used for rotor graphics
    layouts = []
    custom_spritesets = []
    prepare_layout = None
    purchase_prepare_layout = None
    foundations = None

    for graphics in graphics_block.graphics_list:
        cargo_id = graphics.cargo_id
        if isinstance(cargo_id, expression.Identifier):
            cb_name = cargo_id.value
            cb_table = action3_callbacks.callbacks[feature]
            if cb_name in cb_table:
                if cb_name in seen_callbacks:
                    raise generic.ScriptError("Callback '{}' is defined multiple times.".format(cb_name), cargo_id.pos)
                seen_callbacks.add(cb_name)

                info_list = cb_table[cb_name]
                if not isinstance(info_list, list):
                    info_list = [info_list]

                for info in info_list:
                    if "deprecate_message" in info:
                        generic.print_warning(generic.Warning.DEPRECATION, info["deprecate_message"], cargo_id.pos)
                    if house_tile is not None and "tiles" in info and house_tile not in info["tiles"]:
                        continue

                    if info["type"] == "cargo":
                        # Not a callback, but an alias for a certain cargo type
                        if info["num"] in cargo_gfx:
                            raise generic.ScriptError(
                                "Graphics for '{}' are defined multiple times.".format(cb_name), cargo_id.pos
                            )
                        cargo_gfx[info["num"]] = graphics.result.value
                    elif info["type"] == "cb":
                        callbacks.append((info, graphics.result.value))
                    elif info["type"] == "override":
                        assert livery_override is None
                        livery_override = graphics.result.value
                    elif info["type"] == "layout":
                        layouts = graphics.result.value
                        if isinstance(layouts, expression.ConstantNumeric):
                            if layouts.value not in station_sprite_layouts:
                                raise generic.ScriptError("Unknown station", cargo_id.pos)
                            actions = action0.get_copy_layout_action0(feature, id, layouts)
                            layouts = station_sprite_layouts[layouts.value]
                        elif isinstance(layouts, expression.Array) and len(layouts.values) % 2 == 0:
                            actions, var10map, registers_ref = action2layout.parse_station_layouts(
                                feature, id, layouts.values
                            )
                            layouts = (var10map, registers_ref)
                        else:
                            raise generic.ScriptError(
                                "'{}' must be an array of even length, or the ID of a station".format(cb_name),
                                cargo_id.pos,
                            )
                        prepend_action_list.extend(actions)
                        station_sprite_layouts[id.value] = layouts
                    elif info["type"] == "custom_spritesets":
                        if (
                            not isinstance(graphics.result.value, expression.Array)
                            or len(graphics.result.value.values) > 6
                        ):
                            raise generic.ScriptError(
                                "'{}' must be an array of at most 6 elements".format(cb_name), cargo_id.pos
                            )
                        custom_spritesets = graphics.result.value.values
                    elif info["type"] == "prepare_layout":
                        if "purchase" in info:
                            assert purchase_prepare_layout is None
                            purchase_prepare_layout = graphics.result.value
                        else:
                            assert prepare_layout is None
                            prepare_layout = graphics.result.value
                    elif info["type"] == "foundations":
                        assert foundations is None
                        foundations = graphics.result.value
                    else:
                        raise AssertionError()
                continue

        # Not a callback, so it must be a 'normal' cargo (vehicles/stations only)
        cargo_id = cargo_id.reduce_constant(global_constants.const_list)
        # Raise the error only now, to let the 'unknown identifier' take precedence
        if feature >= 5:
            raise generic.ScriptError(
                "Associating graphics with a specific cargo is possible only for vehicles and stations.", cargo_id.pos
            )
        if cargo_id.value in cargo_gfx:
            raise generic.ScriptError(
                "Graphics for cargo {:d} are defined multiple times.".format(cargo_id.value), cargo_id.pos
            )
        cargo_gfx[cargo_id.value] = graphics.result.value

    if graphics_block.default_graphics is not None:
        if "default" not in action3_callbacks.callbacks[feature]:
            raise generic.ScriptError(
                "Default graphics may not be defined for this feature (0x{:02X}).".format(feature),
                graphics_block.default_graphics.pos,
            )
        if None in cargo_gfx:
            raise generic.ScriptError("Default graphics are defined twice.", graphics_block.default_graphics.pos)
        cargo_gfx[None] = graphics_block.default_graphics.value

    if layouts or foundations:
        if None not in cargo_gfx:
            cargo_gfx[None] = expression.SpriteGroupRef(expression.Identifier("CB_FAILED", None), [], None)

        if purchase_prepare_layout and 0xFF not in cargo_gfx:
            cargo_gfx[0xFF] = cargo_gfx[None]

        var10map, registers_ref = layouts or (None, None)

        for cargo in sorted(cargo_gfx, key=lambda x: -1 if x is None else x):
            default = cargo_gfx[cargo]
            varact2parser = action2var.Varaction2Parser(feature)
            # Prepare registers for sprite layout
            if cargo == 0xFF and purchase_prepare_layout:
                varact2parser.parse_expr(purchase_prepare_layout)
                varact2parser.var_list.append(nmlop.VAL2)
                varact2parser.var_list_size += 1
            elif prepare_layout:
                if not isinstance(prepare_layout, expression.SpriteGroupRef):
                    actions, prepare_layout = action2var.create_return_action(
                        prepare_layout, feature, "Station Layout@prepare - Id {:02X}".format(id.value), 0x89
                    )
                    prepend_action_list.extend(actions)
                varact2parser.parse(prepare_layout)
                varact2parser.var_list.append(nmlop.VAL2)
                varact2parser.var_list_size += 1
            # Call the registers procedure
            if registers_ref:
                var_access = action2var.VarAction2ProcCallVar(registers_ref)
                varact2parser.var_list.append(var_access)
                varact2parser.var_list.append(nmlop.VAL2)
                varact2parser.var_list_size += var_access.get_size() + 1
                varact2parser.proc_call_list.append(registers_ref)

            # Fill var10 dependant choices
            mapping = {0x02: (foundations, None)} if foundations else {}
            if var10map:
                var10map.append_mapping(mapping, feature, prepend_action_list, default, custom_spritesets)
            # No need for in-between varaction2 if there are no registers to set and no var10 dependant choices
            if len(mapping) == 0:
                if not varact2parser.var_list:
                    continue
                # mapping must not be empty
                mapping[0x00] = (default, None)
            expr = expression.Variable(expression.ConstantNumeric(0x10), mask=expression.ConstantNumeric(0xFF))
            varact2parser.parse(expr)
            actions, cb_ref = create_intermediate_varaction2(
                feature, varact2parser, mapping, default, graphics_block.pos
            )
            prepend_action_list.extend(actions)
            cargo_gfx[cargo] = cb_ref

    # An in-between varaction2 is always needed for houses
    if len(callbacks) != 0 or feature == 0x07:
        cb_flags = 0
        # Determine the default value
        if None not in cargo_gfx:
            cargo_gfx[None] = expression.SpriteGroupRef(expression.Identifier("CB_FAILED", None), [], None)
        default_val = cargo_gfx[None]

        cb_mapping = {}
        cb_buy_mapping = {}
        # Special case for vehicle cb 36, maps var10 values to spritegroups
        cb36_mapping = {}
        cb36_buy_mapping = {}
        # Special case for industry production CB, maps var18 values to spritegroups
        prod_cb_mapping = {}

        for cb_info, gfx in callbacks:
            if "flag_bit" in cb_info:
                # Set a bit in the CB flags property
                cb_flags |= 1 << cb_info["flag_bit"]

            value_function = cb_info.get("value_function", None)
            mapping_val = (gfx, value_function)

            # See action3_callbacks for info on possible values
            purchase = cb_info.get("purchase", 0)
            if isinstance(purchase, str):
                # Not in purchase list, if separate purchase CB is set
                purchase = 0 if purchase in seen_callbacks else 1
            # Explicit purchase CBs will need a purchase cargo, even if not needed for graphics
            if purchase == 2 and 0xFF not in cargo_gfx:
                cargo_gfx[0xFF] = default_val

            num = cb_info["num"]
            if num == 0x36:
                if purchase != 2:
                    cb36_mapping[cb_info["var10"]] = mapping_val
                if purchase != 0:
                    cb36_buy_mapping[cb_info["var10"]] = mapping_val
            elif feature == 0x0A and num == 0x00:
                # Industry production CB
                assert purchase == 0
                prod_cb_mapping[cb_info["var18"]] = mapping_val
            else:
                if purchase != 2:
                    cb_mapping[num] = mapping_val
                if purchase != 0:
                    cb_buy_mapping[num] = mapping_val

        if cb_flags != 0:
            prepend_action_list.extend(action0.get_callback_flags_actions(feature, id, cb_flags))

        # Handle CB 36
        if len(cb36_mapping) != 0:
            expr = expression.Variable(expression.ConstantNumeric(0x10), mask=expression.ConstantNumeric(0xFF))
            actions, cb36_ref = create_cb_choice_varaction2(
                feature, expr, cb36_mapping, default_val, graphics_block.pos
            )
            prepend_action_list.extend(actions)
            cb_mapping[0x36] = (cb36_ref, None)
        if len(cb36_buy_mapping) != 0:
            expr = expression.Variable(expression.ConstantNumeric(0x10), mask=expression.ConstantNumeric(0xFF))
            actions, cb36_ref = create_cb_choice_varaction2(
                feature, expr, cb36_buy_mapping, default_val, graphics_block.pos
            )
            prepend_action_list.extend(actions)
            cb_buy_mapping[0x36] = (cb36_ref, None)
        if len(prod_cb_mapping) != 0:
            expr = expression.Variable(expression.ConstantNumeric(0x18), mask=expression.ConstantNumeric(0xFF))
            actions, cb_ref = create_cb_choice_varaction2(
                feature, expr, prod_cb_mapping, default_val, graphics_block.pos
            )
            prepend_action_list.extend(actions)
            cb_mapping[0x00] = (cb_ref, None)

        for cargo in sorted(cargo_gfx, key=lambda x: -1 if x is None else x):
            mapping = cb_buy_mapping if cargo == 0xFF else cb_mapping
            if len(mapping) == 0 and feature != 0x07:
                # No callbacks here, so move along
                # Except for houses, where we need to store some stuff in a register
                continue
            if cargo_gfx[cargo] != default_val:
                # There are cargo-specific graphics, be sure to handle those
                # Unhandled callbacks should chain to the default, though
                mapping = mapping.copy()
                mapping[0x00] = (cargo_gfx[cargo], None)

            expr = expression.Variable(expression.ConstantNumeric(0x0C), mask=expression.ConstantNumeric(0xFFFF))
            if feature == 0x07:
                # Store relative x/y, item id (of the north tile) and house tile (HOUSE_TILE_XX constant) in register FF
                # Format: 0xIIHHYYXX: II: item ID, HH: house tile, YY: relative y, XX: relative x
                lowbytes_dict = {
                    "n": 0x000000,
                    "e": 0x010100,
                    "w": 0x020001,
                    "s": 0x030101,
                }
                lowbytes = expression.ConstantNumeric(lowbytes_dict[house_tile])
                highbyte = nmlop.SHIFT_LEFT(house_north_tile_id, 24)
                register_FF = nmlop.OR(lowbytes, highbyte).reduce()
                register_FF = nmlop.STO_TMP(register_FF, 0xFF)
                expr = nmlop.VAL2(register_FF, expr)

                if len(mapping) == 0:
                    # mapping must not be empty
                    mapping[0x00] = (default_val, None)

            actions, cb_ref = create_cb_choice_varaction2(feature, expr, mapping, default_val, graphics_block.pos)
            prepend_action_list.extend(actions)
            cargo_gfx[cargo] = cb_ref

    # Make sure to sort to make the order well-defined
    offset = 7 if feature <= 3 else 5
    for cargo_id in sorted(cg for cg in cargo_gfx if cg is not None):
        result, comment = action2var.parse_result(cargo_gfx[cargo_id], action_list, act6, offset + 1, act3, None, 0x89)
        act3.cid_mappings.append((cargo_id, result, comment))
        offset += 3

    if None in cargo_gfx:
        result, comment = action2var.parse_result(cargo_gfx[None], action_list, act6, offset, act3, None, 0x89)
        act3.def_cid = result
        act3.default_comment = comment
    else:
        act3.def_cid = None
        act3.default_comment = ""

    if livery_override is not None:
        act6livery = action6.Action6()
        # Add any extra actions before the main action3 (TTDP requirement)
        act3livery = create_action3(feature, id, action_list, act6livery, True)
        offset = 7 if feature <= 3 else 5
        result, comment = action2var.parse_result(
            livery_override, action_list, act6livery, offset, act3livery, None, 0x89
        )
        act3livery.def_cid = result
        act3livery.default_comment = comment

    if len(act6.modifications) > 0:
        action_list.append(act6)
    action_list.append(act3)

    if livery_override is not None:
        if len(act6livery.modifications) > 0:
            action_list.append(act6livery)
        action_list.append(act3livery)  # lgtm[py/uninitialized-local-variable]
    action6.free_parameters.restore()

    return prepend_action_list + action_list



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import generic, global_constants
from nml.actions import base_action
from nml.ast import base_statement, general

total_action2_ids = 0x100
free_action2_ids = list(range(0, total_action2_ids))

"""
Statistics about spritegroups.
The 1st field of type C{int} contains the largest number of concurrently active spritegroup ids.
The 2nd field of type L{Position} contains a positional reference
  to the last spritegroup of the concurrently active ones.
"""
spritegroup_stats = (0, None)

total_tmp_locations = 0x7F

"""
Statistics about temporary Action2 registers.
The 1st field of type C{int} contains the largest number of concurrently active register ids.
The 2nd field of type L{Position} contains a positional reference to the spritegroup.
"""
a2register_stats = (0, None)


def print_stats():
    """
    Print statistics about used ids.
    """
    if spritegroup_stats[0] > 0:
        generic.print_info(
            "Concurrent spritegroups: {}/{} ({})".format(
                spritegroup_stats[0], total_action2_ids, str(spritegroup_stats[1])
            )
        )
    if a2register_stats[0] > 0:
        generic.print_info(
            "Concurrent Action2 registers: {}/{} ({})".format(
                a2register_stats[0], total_tmp_locations, str(a2register_stats[1])
            )
        )


class Action2(base_action.BaseAction):
    """
    Abstract Action2 base class.

    @ivar name: Name of the action2.
    @type name: C{str}

    @ivar feature: Action2 feature byte.
    @type feature: C{int}

    @ivar pos: Position reference to source.
    @type pos: L{Position}

    @ivar num_refs: Number of references to this action2.
    @type num_refs: C{int}

    @ivar id: Number of this action2.
    @type id: C{int}, or C{None} if no number is allocated yet.

    @ivar references: All Action2s that are referenced by this Action2.
    @type references: C{list} of L{Action2Reference}

    @ivar tmp_locations: List of address in the temporary storage that are free
                         to be used in this varaction2.
    @type tmp_locations: C{list} of C{int}
    """

    def __init__(self, feature, name, pos):
        self.feature = feature
        self.name = name
        self.pos = pos
        self.num_refs = 0
        self.id = None
        self.references = []
        # 0x00 - 0x7F: available to user
        # 0x80 - 0xFE: used by NML
        # 0xFF: Used for some house variables
        # 0x100 - 0x10F: Special meaning (used for some CB results)
        self.tmp_locations = list(range(0x80, 0x80 + total_tmp_locations))

    def prepare_output(self, sprite_num):
        free_references(self)

        global spritegroup_stats

        try:
            if self.num_refs == 0:
                self.id = free_action2_ids[0]
            else:
                self.id = free_action2_ids.pop()
                num_used = total_action2_ids - len(free_action2_ids)
                if num_used > spritegroup_stats[0]:
                    spritegroup_stats = (num_used, self.pos)
        except IndexError:
            raise generic.ScriptError(
                "Unable to allocate ID for [random]switch, sprite set/layout/group or produce-block."
                " Try reducing the number of such blocks.",
                self.pos,
            )

    def write_sprite_start(self, file, size, extra_comment=None):
        assert self.num_refs == 0, "Action2 reference counting has {:d} dangling references.".format(self.num_refs)
        file.comment("Name: " + self.name)
        if extra_comment:
            for c in extra_comment:
                file.comment(c)
        file.start_sprite(size + 3)
        file.print_bytex(2)
        file.print_bytex(self.feature)
        file.print_bytex(self.id)

    def skip_action7(self):
        return False

    def skip_action9(self):
        return False

    def skip_needed(self):
        return False

    def remove_tmp_location(self, location, force_recursive):
        """
        Recursively remove a location from the list of available temporary
        storage locations. It is not only removed from the list of the
        current Action2Var but also from all Action2Var it calls. If an
        Action2Var is referenced as a procedure call, the location is always
        removed recursively, otherwise only if force_recursive is True.

        @param location: Number of the storage register to remove.
        @type location: C{int}

        @param force_recursive: Force removing this location recursively,
                                also for 'chained' action2s.
        @type force_recursive: C{bool}
        """
        global a2register_stats

        if location in self.tmp_locations:
            self.tmp_locations.remove(location)
            num_used = total_tmp_locations - len(self.tmp_locations)
            if num_used > a2register_stats[0]:
                a2register_stats = (num_used, self.pos)

        for act2_ref in self.references:
            if force_recursive or act2_ref.is_proc:
                act2_ref.action2.remove_tmp_location(location, True)


class Action2Reference:
    """
    Container class to store information about an action2 reference

    @ivar action2: The target action2
    @type action2: L{Action2}

    @ivar is_proc: Whether this reference is made because of a procedure call
    @type is_proc: C{bool}
    """

    def __init__(self, action2, is_proc):
        self.action2 = action2
        self.is_proc = is_proc


def add_ref(ref, source_action, reference_as_proc=False):
    """
    Add a reference to a certain action2.
    This is needed so we can correctly reserve / free action2 IDs later on.
    To be called when creating the actions from the AST.

    @param ref: Reference to the sprite group that corresponds to the action2.
    @type ref: L{SpriteGroupRef}

    @param source_action: Source action (act2 or act3) that contains the reference.
    @type source_action: L{Action2} or L{Action3}

    @param reference_as_proc: True iff the reference source is a procedure call,
                              which needs special precautions for temp registers.
    @type reference_as_proc: C{bool}
    """

    # Add reference to list of references of the source action
    act2 = ref.act2 if ref.act2 is not None else resolve_spritegroup(ref.name).get_action2(source_action.feature)
    source_action.references.append(Action2Reference(act2, reference_as_proc))
    act2.num_refs += 1


def free_references(source_action):
    """
    Free all references to other action2s from a certain action 2/3

    @param source_action: Action that contains the reference
    @type  source_action: L{Action2} or L{Action3}
    """
    for act2_ref in source_action.references:
        act2 = act2_ref.action2
        act2.num_refs -= 1
        if act2.num_refs == 0:
            free_action2_ids.append(act2.id)


# Features using sprite groups directly: vehicles, stations, canals, cargos, railtypes, airports, roadtypes, tramtypes
features_sprite_group = [0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x0B, 0x0D, 0x10, 0x12, 0x13]
# Features using sprite layouts: houses, industry tiles, objects, airport tiles, and road stops
features_sprite_layout = [0x07, 0x09, 0x0F, 0x11, 0x14]
# All features that need sprite sets
features_sprite_set = features_sprite_group + features_sprite_layout


def make_sprite_group_class(cls_is_spriteset, cls_is_referenced, cls_has_explicit_feature, cls_is_relocatable=False):
    """
    Metaclass factory which makes base classes for all nodes 'Action 2 graph'
    This graph is made up of all blocks that are eventually compiled to Action2,
    which use the same name space. Spritesets do inherit from this class to make
    referencing them possible, but they are not part of the reference graph that
    is built.

    @param cls_is_spriteset: Whether this class represents a spriteset
    @type cls_is_spriteset: C{bool}

    @param cls_is_referenced: True iff this node can be referenced by other nodes
    @type cls_is_referenced: C{bool}

    @param cls_has_explicit_feature: Whether the feature of an instance is explicitly set,
                                    or derived from nodes that link to it.
    @type cls_has_explicit_feature: C{bool}

    @param cls_is_relocatable: Whether instances of this class can be freely moved around or whether they need
                               to be converted to nfo code at the same location as they are in the nml code.
    @type cls_is_relocatable: C{bool}

    @return: The constructed class
    @rtype: C{type}
    """

    # without either references or an explicit feature, we have nothing to base our feature on
    assert cls_is_referenced or cls_has_explicit_feature

    class ASTSpriteGroup(base_statement.BaseStatement):
        """
        Abstract base class for all AST nodes that represent a sprite group
        This handles all the relations between the various nodes

        Child classes should do the following:
            - Implement their own __init__ method
            - Call BaseStatement.__init__
            - Call initialize, pre_process and prepare_output (in that order)
            - Implement collect_references
            - Call set_action2 after generating the corresponding action2 (if applicable)

        @ivar _referencing_nodes: Set of nodes that refer to this node
        @type _referencing_nodes: C{set}

        @ivar _referenced_nodes: Set of nodes that this node refers to
        @type _referenced_nodes: C{set}

        @ivar _prepared: True iff prepare_output has already been executed
        @type _prepared: C{bool}

        @ivar _action2: Mapping of features to action2s
        @type _action2: C{dict} that maps C{int} to L{Action2}

        @ivar feature_set: Set of features that use this node
        @type feature_set: C{set} of C{int}

        @ivar name: Name of this node, as declared by the user
        @type name: L{Identifier}

        @ivar num_params: Number of parameters that can be (and have to be) passed
        @type num_params: C{int}

        @ivar used_sprite_sets: List of sprite sets used by this node
        @type used_sprite_sets: C{list} of L{SpriteSet}
        """

        def __init__(self):
            """
            Subclasses should implement their own __init__ method.
            This method should not be called, because calling a method on a metaclass can be troublesome.
            Instead, call initialize(...).
            """
            raise NotImplementedError(
                (
                    "__init__ must be implemented in ASTSpriteGroup-subclass {!r},"
                    " initialize(..) should be called instead"
                ).format(type(self))
            )

        def initialize(self, name=None, feature=None, num_params=0):
            """
            Initialize this instance.
            This function is generally, but not necessarily, called from the child class' constructor.
            Calling it later (during pre-processing) is also possible, as long as it's called
            before any other actions are done.

            @param name: Name of this node, as set by the user (if applicable)
                            Should be be set (not None) iff cls_is_referenced is True
            @type name: L{Identifier} or C{None} if N/A

            @param feature: Feature of this node, if set by the user.
                                Should be set (not None) iff cls_has_explicit_feature is True
            @type feature: C{int} or C{None}
            """
            assert not (self._has_explicit_feature() and feature is None)
            assert not (cls_is_referenced and name is None)
            self._referencing_nodes = set()
            self._referenced_nodes = set()
            self._prepared = False
            self._action2 = {}
            self.feature_set = {feature} if feature is not None else set()
            self.name = name
            self.num_params = num_params
            self.used_sprite_sets = []
            self.optimised = None

        def register_names(self):
            if cls_is_relocatable and cls_is_referenced:
                register_spritegroup(self)

        def pre_process(self):
            """
            Pre-process this node.
            During this stage, the reference graph is built.
            """
            refs = self.collect_references()
            for ref in refs:
                self._add_reference(ref)
            if (not cls_is_relocatable) and cls_is_referenced:
                register_spritegroup(self)

        def prepare_act2_output(self):
            """
            Prepare this node for outputting.
            This sets the feature and makes sure it is correct.

            @return: True iff parsing of this node is needed
            @rtype: C{bool}
            """
            if not cls_is_referenced:
                return True
            if not self._prepared:
                self._prepared = True
                # copy, since we're going to modify
                ref_nodes = self._referencing_nodes.copy()
                for node in ref_nodes:
                    used = node.prepare_act2_output()
                    if not used:
                        node._remove_reference(self)

                # now determine the feature
                if self._has_explicit_feature():
                    # by this time, feature should be set
                    assert len(self.feature_set) == 1
                    for n in self._referencing_nodes:
                        if n.feature_set != self.feature_set:
                            msg = "Cannot refer to block '{}' with feature '{}', expected feature is '{}'"
                            msg = msg.format(
                                self.name.value,
                                general.feature_name(next(iter(self.feature_set))),
                                general.feature_name(n.feature_set.difference(self.feature_set).pop()),
                            )
                            raise generic.ScriptError(msg, n.pos)

                elif len(self._referencing_nodes) != 0:
                    for n in self._referencing_nodes:
                        # Add the features from all calling blocks to the set
                        self.feature_set.update(n.feature_set)

                if len(self._referencing_nodes) == 0 and (not self.optimised or self.optimised is self):
                    # if we can be 'not used', there ought to be a way to refer to this block
                    assert self.name is not None
                    generic.print_warning(
                        generic.Warning.OPTIMISATION,
                        "Block '{}' is not referenced, ignoring.".format(self.name.value),
                        self.pos,
                    )

            return len(self._referencing_nodes) != 0

        def referenced_nodes(self):
            """
            Get the nodes that this node refers to.
            @note: Make sure to sort this in a deterministic way when the order of items affects the output.

            @return: A set of nodes
            @rtype: C{set} of L{ASTSpriteGroup}
            """
            return self._referenced_nodes

        def referencing_nodes(self):
            """
            Get the nodes that refer to this node.
            @note: Make sure to sort this in a deterministic way when the order of items affects the output.

            @return: A set of nodes
            @rtype: C{set} of L{ASTSpriteGroup}
            """

            return self._referencing_nodes

        def optimise(self):
            """
            Optimise this sprite group.

            @return: True iff this sprite group has been optimised
            @rtype: C{bool}
            """
            return False

        def collect_references(self):
            """
            This function should collect all references to other nodes from this instance.

            @return: A collection containing all links to other nodes.
            @rtype: C{iterable} of L{SpriteGroupRef}
            """
            raise NotImplementedError(
                "collect_references must be implemented in ASTSpriteGroup-subclass {!r}".format(type(self))
            )

        def set_action2(self, action2, feature):
            """
            Set this node's resulting action2

            @param feature: Feature of the Action2
            @type feature: C{int}

            @param action2: Action2 to set
            @type action2: L{Action2}
            """
            assert feature not in self._action2
            self._action2[feature] = action2

        def get_action2(self, feature):
            """
            Get this node's resulting action2

            @param feature: Feature of the Action2
            @type feature: C{int}

            @return: Action2 to get
            @rtype: L{Action2}
            """
            assert feature in self._action2
            return self._action2[feature]

        def has_action2(self, feature):
            """
            Check, if this node already has an action2 for a given feature

            @param feature: Feature to check
            @type feature: C{int}

            @return: True iff there is an action2 for this feature
            @rtype: C{bool}
            """
            return feature in self._action2

        def _add_reference(self, target_ref):
            """
            Add a reference from C{self} to a target with a given name.

            @param target_ref: Name of the reference target
            @type target_ref: L{SpriteGroupRef}
            """

            if target_ref.name.value == "CB_FAILED":
                return

            target = resolve_spritegroup(target_ref.name)
            if target.is_spriteset():
                assert target.num_params == 0
                # Referencing a spriteset directly from graphics/[random]switch
                # Passing parameters is not possible here
                if len(target_ref.param_list) != 0:
                    raise generic.ScriptError(
                        "Passing parameters to '{}' is only possible from a spritelayout.".format(
                            target_ref.name.value
                        ),
                        target_ref.pos,
                    )

                self.used_sprite_sets.append(target)
            else:
                if len(target_ref.param_list) != target.num_params:
                    msg = "'{}' expects {:d} parameters, encountered {:d}."
                    msg = msg.format(target_ref.name.value, target.num_params, len(target_ref.param_list))
                    raise generic.ScriptError(msg, target_ref.pos)

                self._referenced_nodes.add(target)
                target._referencing_nodes.add(self)

        def _remove_reference(self, target):
            """
            Add a reference from C{self} to a target

            @param target: Existing reference target to be removed
            @type target: L{ASTSpriteGroup}
            """
            assert target in self._referenced_nodes
            assert self in target._referencing_nodes
            self._referenced_nodes.remove(target)
            target._referencing_nodes.remove(self)

        # Make metaclass arguments available outside of the class
        def is_spriteset(self):
            return cls_is_spriteset

        def _has_explicit_feature(self):
            return cls_has_explicit_feature

    return ASTSpriteGroup


# list of all registered sprite sets and sprite groups
spritegroup_list = {}


def register_spritegroup(spritegroup):
    """
    Register a sprite group, so it can be resolved by name later

    @param spritegroup: Sprite group to register
    @type spritegroup: L{ASTSpriteGroup}
    """
    name = spritegroup.name.value
    if name in spritegroup_list:
        raise generic.ScriptError("Block with name '{}' has already been defined".format(name), spritegroup.pos)
    spritegroup_list[name] = spritegroup
    global_constants.spritegroups[name] = name


def resolve_spritegroup(name):
    """
    Resolve a sprite group with a given name

    @param name: Name of the sprite group.
    @type name: L{Identifier}

    @return: The sprite group that the name refers to.
    @rtype: L{ASTSpriteGroup}
    """
    if name.value not in spritegroup_list:
        raise generic.ScriptError("Unknown identifier encountered: '{}'".format(name.value), name.pos)
    return spritegroup_list[name.value]



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import generic, grfstrings

from .base_expression import Expression, Type


class StringLiteral(Expression):
    """
    String literal expression.

    @ivar value: Value of the string literal.
    @type value: C{str}
    """

    def __init__(self, value, pos):
        Expression.__init__(self, pos)
        self.value = value

    def debug_print(self, indentation):
        generic.print_dbg(indentation, 'String literal: "{}"'.format(self.value))

    def __str__(self):
        return '"{}"'.format(self.value)

    def write(self, file, size):
        assert grfstrings.get_string_size(self.value, False, True) == size
        file.print_string(self.value, final_zero=False, force_ascii=True)

    def reduce(self, id_dicts=None, unknown_id_fatal=True):
        return self

    def type(self):
        return Type.STRING_LITERAL



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import expression, generic, global_constants
from nml.actions import action0
from nml.ast import base_statement, general


class SortVehicles(base_statement.BaseStatement):
    """
    AST-node representing a sort-vehicles block.

    @ivar feature: Feature of the item
    @type feature: L{ConstantNumeric}

    @ivar vehid_list: List of vehicle ids.
    @type vehid_list: L{Array}.
    """

    def __init__(self, params, pos):
        base_statement.BaseStatement.__init__(self, "sort-block", pos)
        if len(params) != 2:
            raise generic.ScriptError(
                "Sort-block requires exactly two parameters, got {:d}".format(len(params)), self.pos
            )
        self.feature = general.parse_feature(params[0])
        self.vehid_list = params[1]

    def pre_process(self):
        self.vehid_list = self.vehid_list.reduce(global_constants.const_list)
        if not isinstance(self.vehid_list, expression.Array) or not all(
            isinstance(x, expression.ConstantNumeric) for x in self.vehid_list.values
        ):
            raise generic.ScriptError(
                "Second parameter is not an array of one of the items in it could not be reduced to a constant number",
                self.pos,
            )

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Sort, feature", hex(self.feature.value))
        for id in self.vehid_list.values:
            generic.print_dbg(indentation + 2, "Vehicle id:", id)

    def get_action_list(self):
        return action0.parse_sort_block(self.feature.value, self.vehid_list.values)

    def __str__(self):
        return "sort({:d}, {});\n".format(self.feature.value, self.vehid_list)



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import generic
from nml.actions import action7
from nml.ast import base_statement


class SkipAll(base_statement.BaseStatement):
    """
    Skip everything after this statement.
    """

    def __init__(self, pos):
        base_statement.BaseStatement.__init__(self, "exit-statement", pos)

    def get_action_list(self):
        return [action7.UnconditionalSkipAction(9, 0)]

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Skip all")

    def __str__(self):
        return "exit;\n"



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

import calendar
import datetime
import math
from functools import reduce

from nml import generic, global_constants, nmlop

from . import identifier
from .base_expression import ConstantFloat, ConstantNumeric, Expression, Type
from .bitmask import BitMask
from .cargo import AcceptCargo, ProduceCargo
from .parameter import parse_string_to_dword
from .storage_op import StorageOp
from .string_literal import StringLiteral
from .abs_op import AbsOp


class FunctionCall(Expression):
    def __init__(self, name, params, pos):
        Expression.__init__(self, pos)
        self.name = name
        self.params = params

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Call function: " + self.name.value)
        for param in self.params:
            generic.print_dbg(indentation + 2, "Parameter:")
            param.debug_print(indentation + 4)

    def __str__(self):
        ret = "{}({})".format(self.name, ", ".join(str(param) for param in self.params))
        return ret

    def reduce(self, id_dicts=None, unknown_id_fatal=True):
        # At this point we don't care about invalid arguments, they'll be handled later.
        identifier.ignore_all_invalid_ids = True
        params = [param.reduce(id_dicts, unknown_id_fatal=False) for param in self.params]
        identifier.ignore_all_invalid_ids = False
        if self.name.value in function_table:
            func = function_table[self.name.value]
            val = func(self.name.value, params, self.pos)
            return val.reduce(id_dicts)
        else:
            # try user-defined functions
            func_ptr = self.name.reduce(id_dicts, unknown_id_fatal=False, search_func_ptr=True)
            if func_ptr != self.name:  # we found something!
                if func_ptr.type() == Type.SPRITEGROUP_REF:
                    func_ptr.param_list = params
                    func_ptr.is_procedure = True
                    return func_ptr
                if func_ptr.type() != Type.FUNCTION_PTR:
                    raise generic.ScriptError(
                        "'{}' is defined, but it is not a function.".format(self.name.value), self.pos
                    )
                return func_ptr.call(params)
            if unknown_id_fatal:
                raise generic.ScriptError("'{}' is not defined as a function.".format(self.name.value), self.pos)
            return FunctionCall(self.name, params, self.pos)


class SpecialCheck(Expression):
    """
    Action7/9 special check (e.g. to see whether a cargo is defined)

    @ivar op: Action7/9 operator to use
    @type op: (C{int}, C{str})-tuple

    @ivar varnum: Variable number to read
    @type varnum: C{int}

    @ivar results: Result of the check when skipping (0) or not skipping (1)
    @type results: (C{int}, C{int})-tuple

    @ivar value: Value to test
    @type value: C{int}

    @ivar varsize: Varsize for the action7/9 check
    @type varsize: C{int}

    @ivar mask: Mask to test only certain bits of the value
    @type mask: C{int}

    @ivar pos: Position information
    @type pos: L{Position}
    """

    def __init__(self, op, varnum, results, value, to_string, varsize=4, mask=None, pos=None):
        Expression.__init__(self, pos)
        self.op = op
        self.varnum = varnum
        self.results = results
        self.value = value
        self.to_string = to_string
        self.varsize = varsize
        self.mask = mask

    def reduce(self, id_dicts=None, unknown_id_fatal=True):
        return self

    def __str__(self):
        return self.to_string

    def supported_by_actionD(self, raise_error):
        return True


class GRMOp(Expression):
    def __init__(self, op, feature, count, to_string, pos=None):
        Expression.__init__(self, pos)
        self.op = op
        self.feature = feature
        self.count = count
        self.to_string = to_string

    def reduce(self, id_dicts=None, unknown_id_fatal=True):
        return self

    def __str__(self):
        return self.to_string(self)

    def supported_by_actionD(self, raise_error):
        return True


function_table = {}


def builtin(func):
    """
    Decorator that adds a function named `builtin_func` to the function table as `func`.
    """
    assert func.__name__.startswith("builtin_")
    name = func.__name__[8:]  # Strip the "builtin_". str.removeprefix() is only added in py3.9.
    function_table[name] = func
    return func


def builtins(*names):
    """
    Decorator that adds a function to the function table with one or more custom names.
    """

    def dec(func):
        for name in names:
            function_table[name] = func
        return func

    return dec


# { Builtin functions


@builtin
def builtin_min(name, args, pos):
    """
    min(...) builtin function.

    @return Lowest value of the given arguments.
    """
    if len(args) < 2:
        raise generic.ScriptError("min() requires at least 2 arguments", pos)
    return reduce(lambda x, y: nmlop.MIN(x, y, pos), args)


@builtin
def builtin_max(name, args, pos):
    """
    max(...) builtin function.

    @return Highest value of the given arguments.
    """
    if len(args) < 2:
        raise generic.ScriptError("max() requires at least 2 arguments", pos)
    return reduce(lambda x, y: nmlop.MAX(x, y, pos), args)


@builtin
def builtin_date(name, args, pos):
    """
    date(year, month, day) builtin function.

    @return Days since 1 jan 1 of the given date.
    """
    days_in_month = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    if len(args) != 3:
        raise generic.ScriptError("date() requires exactly 3 arguments", pos)
    identifier.ignore_all_invalid_ids = True
    year = args[0].reduce(global_constants.const_list)
    identifier.ignore_all_invalid_ids = False
    try:
        month = args[1].reduce_constant().value
        day = args[2].reduce_constant().value
    except generic.ConstError:
        raise generic.ScriptError("Month and day parameters of date() should be compile-time constants", pos)
    generic.check_range(month, 1, 12, "month", args[1].pos)
    generic.check_range(day, 1, days_in_month[month - 1], "day", args[2].pos)

    if not isinstance(year, ConstantNumeric):
        if month != 1 or day != 1:
            raise generic.ScriptError(
                "when the year parameter of date() is not a compile time constant month and day should be 1", pos
            )
        # num_days = year*365 + year/4 - year/100 + year/400
        part1 = nmlop.MUL(year, 365)
        part2 = nmlop.DIV(year, 4)
        part3 = nmlop.DIV(year, 100)
        part4 = nmlop.DIV(year, 400)
        res = nmlop.ADD(part1, part2)
        res = nmlop.SUB(res, part3)
        res = nmlop.ADD(res, part4)
        return res

    generic.check_range(year.value, 0, 5000000, "year", year.pos)
    day_in_year = 0
    for i in range(month - 1):
        day_in_year += days_in_month[i]
    day_in_year += day
    if month >= 3 and (year.value % 4 == 0) and ((not year.value % 100 == 0) or (year.value % 400 == 0)):
        day_in_year += 1
    return ConstantNumeric(year.value * 365 + calendar.leapdays(0, year.value) + day_in_year - 1, pos)


@builtin
def builtin_day_of_year(name, args, pos):
    """
    day_of_year(month, day) builtin function.

    @return Day of the year, assuming February has 28 days.
    """
    if len(args) != 2:
        raise generic.ScriptError(name + "() must have a month and a day parameter", pos)

    month = args[0].reduce()
    if not isinstance(month, ConstantNumeric):
        raise generic.ScriptError("Month should be a compile-time constant.", month.pos)
    if month.value < 1 or month.value > 12:
        raise generic.ScriptError("Month should be a value between 1 and 12.", month.pos)

    day = args[1].reduce()
    if not isinstance(day, ConstantNumeric):
        raise generic.ScriptError("Day should be a compile-time constant.", day.pos)

    # Mapping of month to number of days in that month.
    number_days = {1: 31, 2: 28, 3: 31, 4: 30, 5: 31, 6: 30, 7: 31, 8: 31, 9: 30, 10: 31, 11: 30, 12: 31}
    if day.value < 1 or day.value > number_days[month.value]:
        raise generic.ScriptError("Day should be value between 1 and {:d}.".format(number_days[month.value]), day.pos)

    return ConstantNumeric(datetime.date(1, month.value, day.value).toordinal(), pos)


@builtins("STORE_TEMP", "STORE_PERM", "LOAD_TEMP", "LOAD_PERM")
def builtin_storage(name, args, pos):
    """
    Accesses to temporary / persistent storage
    """
    return StorageOp(name, args, pos)


@builtin
def builtin_UCMP(name, args, pos):
    if len(args) != 2:
        raise generic.ScriptError(name + "() must have exactly two parameters", pos)
    return nmlop.VACT2_UCMP(args[0], args[1], pos)


@builtin
def builtin_CMP(name, args, pos):
    if len(args) != 2:
        raise generic.ScriptError(name + "() must have exactly two parameters", pos)
    return nmlop.VACT2_CMP(args[0], args[1], pos)


@builtin
def builtin_rotate(name, args, pos):
    if len(args) != 2:
        raise generic.ScriptError(name + "() must have exactly two parameters", pos)
    return nmlop.ROT_RIGHT(args[0], args[1], pos)


@builtin
def builtin_bitmask(name, args, pos):
    return BitMask(args, pos)


@builtin
def builtin_hasbit(name, args, pos):
    """
    hasbit(value, bit_num) builtin function.

    @return C{1} if and only if C{value} has bit C{bit_num} set, C{0} otherwise.
    """
    if len(args) != 2:
        raise generic.ScriptError(name + "() must have exactly two parameters", pos)
    return nmlop.HASBIT(args[0], args[1], pos)


@builtin
def builtin_getbits(name, args, pos):
    """
    getbits(value, first, amount) builtin function.

    @return Extract C{amount} bits starting at C{first} from C{value},
            that is (C{value} >> C{first}) & (1 << C{amount} - 1)
    """
    if len(args) != 3:
        raise generic.ScriptError(name + "() must have exactly three parameters", pos)

    # getbits(value, first, amount) = (value >> first) & ((0xFFFFFFFF << amount) ^ 0xFFFFFFFF)
    part1 = nmlop.SHIFTU_RIGHT(args[0], args[1], pos)
    part2 = nmlop.SHIFT_LEFT(0xFFFFFFFF, args[2], pos)
    part3 = nmlop.XOR(part2, 0xFFFFFFFF, pos)

    return nmlop.AND(part1, part3, pos)


@builtin
def builtin_version_openttd(name, args, pos):
    """
    version_openttd(major, minor[, revision[, build]]) builtin function.

    For OpenTTD >= 12.0 only 2 parameters may be passed.
    For OpenTTD < 12.0 3 to 4 parameters may be passed.

    @return The version information encoded in a double-word.
    """
    if len(args) < 2:
        raise generic.ScriptError(name + "() must have at least 2 parameters", pos)

    major = args[0].reduce_constant().value
    minor = args[1].reduce_constant().value

    if major >= 12:
        if len(args) != 2:
            raise generic.ScriptError(name + "() must have at exactly 2 parameters for OpenTTD >= 12.0", pos)

        return ConstantNumeric(((16 + major) << 24) | (minor << 20))
    else:
        if len(args) > 4:
            raise generic.ScriptError(name + "() must have at most 4 parameters", pos)

        revision = args[2].reduce_constant().value if len(args) >= 3 else 0
        build = args[3].reduce_constant().value if len(args) >= 4 else 0x80000
        return ConstantNumeric((major << 28) | (minor << 24) | (revision << 20) | build)


@builtins("cargotype_available", "railtype_available", "roadtype_available", "tramtype_available")
def builtin_typelabel_available(name, args, pos):
    """
    {cargo|rail|road|tram}type_available(label) builtin functions.

    @return 1 if the label is available, 0 otherwise.
    """
    op = {
        "cargotype_available": (0x0B, r"\7c"),
        "railtype_available": (0x0D, None),
        "roadtype_available": (0x0F, None),
        "tramtype_available": (0x11, None),
    }[name]

    if len(args) != 1:
        raise generic.ScriptError(name + "() must have exactly 1 parameter", pos)
    label = args[0].reduce()
    return SpecialCheck(op, 0, (0, 1), parse_string_to_dword(label), "{}({})".format(name, label), pos=args[0].pos)


@builtins("grf_current_status", "grf_future_status", "grf_order_behind")
def builtin_grf_status(name, args, pos):
    """
    grf_{current_status|future_status|order_behind}(grfid[, mask]) builtin functions.

    @return 1 if the grf is, or will be, active, 0 otherwise.
    """
    op, results = {
        # can't use \7g (0, 1), because that's false when the queried grf isn't present at all.
        "grf_current_status": ((0x06, r"\7G"), (1, 0)),
        "grf_future_status": ((0x0A, r"\7gg"), (0, 1)),
        "grf_order_behind": ((0x08, r"\7gG"), (0, 1)),
    }[name]

    if len(args) == 1:
        grfid = args[0].reduce()
        mask = None
        string = "{}({})".format(name, grfid)
        varsize = 4
    elif len(args) == 2:
        grfid = args[0].reduce()
        mask = parse_string_to_dword(args[1].reduce())
        string = "{}({}, {})".format(name, grfid, mask)
        varsize = 8
    else:
        raise generic.ScriptError(name + "() must have 1 or 2 parameters", pos)

    return SpecialCheck(op, 0x88, results, parse_string_to_dword(grfid), string, varsize, mask, args[0].pos)


@builtins("visual_effect", "visual_effect_and_powered")
def builtin_visual_effect_and_powered(name, args, pos):
    """
    Builtin function, used in two forms:
    visual_effect_and_powered(effect, offset, powered)
    visual_effect(effect, offset)
    Use this to set the vehicle property visual_effect[_and_powered]
    and for the callback VEH_CB_VISUAL_EFFECT[_AND_POWERED]

    """
    arg_len = 2 if name == "visual_effect" else 3
    if len(args) != arg_len:
        raise generic.ScriptError(name + "() must have {:d} parameters".format(arg_len), pos)
    effect = args[0].reduce_constant(global_constants.const_list).value
    offset = nmlop.ADD(args[1], 8).reduce_constant().value
    generic.check_range(offset, 0, 0x0F, "offset in function " + name, pos)
    if arg_len == 3:
        powered = args[2].reduce_constant(global_constants.const_list).value
        if powered != 0 and powered != 0x80:
            raise generic.ScriptError(
                "3rd argument to visual_effect_and_powered (powered) must be"
                " either ENABLE_WAGON_POWER or DISABLE_WAGON_POWER",
                pos,
            )
    else:
        powered = 0
    return ConstantNumeric(effect | offset | powered)


@builtin
def builtin_create_effect(name, args, pos):
    """
    Builtin function:
    create_effect(effect_sprite, l_x_offset, t_y_offset, z_offset)
    Use this to set the values for temporary storages 100+x
    in the callback create_effect

    """
    if len(args) != 4:
        raise generic.ScriptError(name + "() must have 4 parameters", pos)

    sprite = args[0].reduce_constant(global_constants.const_list).value
    offset1 = args[1].reduce_constant().value
    offset2 = args[2].reduce_constant().value
    offset3 = args[3].reduce_constant().value

    generic.check_range(sprite, 0, 255, "effect_sprite in function " + name, args[0].pos)
    generic.check_range(offset1, -128, 127, "l_x_offset in function " + name, args[1].pos)
    generic.check_range(offset2, -128, 127, "t_y_offset in function " + name, args[2].pos)
    generic.check_range(offset3, -128, 127, "z_offset in function " + name, args[3].pos)

    return ConstantNumeric(sprite | (offset1 & 0xFF) << 8 | (offset2 & 0xFF) << 16 | (offset3 & 0xFF) << 24)


@builtin
def builtin_str2number(name, args, pos):
    if len(args) != 1:
        raise generic.ScriptError(name + "() must have 1 parameter", pos)
    return ConstantNumeric(parse_string_to_dword(args[0]))


@builtins("cargotype", "railtype", "roadtype", "tramtype")
def builtin_resolve_typelabel(name, args, pos, table_name=None):
    """
    {cargo,rail,road,tram}type(label) builtin functions.

    Also used from some Action2Var variables to resolve cargo labels.
    """
    tracktype_funcs = {
        "cargotype": global_constants.cargo_numbers,
        "railtype": global_constants.railtype_table,
        "roadtype": global_constants.roadtype_table,
        "tramtype": global_constants.tramtype_table,
    }

    if not table_name:
        table_name = name
    table = tracktype_funcs[table_name]
    if table_name == "cargotype":
        table_name = "cargo"  # NML syntax uses "cargotable" and "railtypetable"

    if len(args) != 1:
        raise generic.ScriptError(name + "() must have 1 parameter", pos)
    if not isinstance(args[0], StringLiteral) or args[0].value not in table:
        raise generic.ScriptError(
            "Parameter for {}() must be a string literal that is also in your {} table".format(name, table_name), pos
        )
    return ConstantNumeric(table[args[0].value])


@builtin
def builtin_reserve_sprites(name, args, pos):
    if len(args) != 1:
        raise generic.ScriptError(name + "() must have 1 parameter", pos)
    count = args[0].reduce_constant()
    return GRMOp(nmlop.GRM_RESERVE, 0x08, count.value, lambda x: "{}({:d})".format(name, count.value), pos)


@builtin
def builtin_industry_type(name, args, pos):
    """
    industry_type(IND_TYPE_OLD | IND_TYPE_NEW, id) builtin function

    @return The industry type in the format used by grfs (industry prop 0x16 and var 0x64)
    """
    if len(args) != 2:
        raise generic.ScriptError(name + "() must have 2 parameters", pos)

    type = args[0].reduce_constant(global_constants.const_list).value
    if type not in (0, 1):
        raise generic.ScriptError("First argument of industry_type() must be IND_TYPE_OLD or IND_TYPE_NEW", pos)

    # Industry ID uses 7 bits (0 .. 6), bit 7 is for old/new
    id = args[1].reduce_constant(global_constants.const_list).value
    if not 0 <= id <= 127:
        raise generic.ScriptError("Second argument 'id' of industry_type() must be in range 0..127", pos)

    return ConstantNumeric(type << 7 | id)


@builtins("accept_cargo", "produce_cargo")
def builtin_cargoexpr(name, args, pos):
    if len(args) < 1:
        raise generic.ScriptError(name + "() must have 1 or more parameters", pos)

    if not isinstance(args[0], StringLiteral) or args[0].value not in global_constants.cargo_numbers:
        raise generic.ScriptError(
            "First argument of " + name + "() must be a string literal that is also in your cargo table", pos
        )
    cargotype = global_constants.cargo_numbers[args[0].value]

    if name == "produce_cargo":
        return ProduceCargo(cargotype, args[1:], pos)
    elif name == "accept_cargo":
        return AcceptCargo(cargotype, args[1:], pos)
    else:
        raise AssertionError()


@builtins("acos", "asin", "atan", "cos", "sin", "sqrt", "tan")
def builtin_math(name, args, pos):
    if len(args) != 1:
        raise generic.ScriptError(name + "() must have 1 parameter", pos)
    val = args[0].reduce()
    if not isinstance(val, (ConstantNumeric, ConstantFloat)):
        raise generic.ScriptError("Parameter for " + name + "() must be a constant", pos)
    math_func_table = {
        "acos": math.acos,
        "asin": math.asin,
        "atan": math.atan,
        "cos": math.cos,
        "sin": math.sin,
        "sqrt": math.sqrt,
        "tan": math.tan,
    }
    return ConstantFloat(math_func_table[name](val.value), val.pos)


@builtin
def builtin_round(name, args, pos):
    if len(args) != 1:
        raise generic.ScriptError(name + "() must have 1 parameter", pos)
    val = args[0].reduce()
    if not isinstance(val, (ConstantNumeric, ConstantFloat)):
        raise generic.ScriptError("Parameter for " + name + "() must be a constant", pos)
    return ConstantNumeric(round(val.value), pos)


@builtin
def builtin_int(name, args, pos):
    if len(args) != 1:
        raise generic.ScriptError(name + "() must have 1 parameter", pos)
    val = args[0].reduce()
    if not isinstance(val, (ConstantNumeric, ConstantFloat)):
        raise generic.ScriptError("Parameter for " + name + "() must be a constant", pos)
    return ConstantNumeric(int(val.value), val.pos)


@builtin
def builtin_abs(name, args, pos):
    if len(args) != 1:
        raise generic.ScriptError(name + "() must have 1 parameter", pos)
    return AbsOp(args[0], args[0].pos).reduce()


@builtin
def builtin_sound(name, args, pos):
    if len(args) not in (1, 2):
        raise generic.ScriptError(name + "() must have 1 or 2 parameters", pos)
    if not isinstance(args[0], StringLiteral):
        raise generic.ScriptError("Parameter for " + name + "() must be a string literal", pos)
    volume = args[1].reduce_constant().value if len(args) >= 2 else 100
    generic.check_range(volume, 0, 100, "sound volume", pos)
    from nml.actions import action11

    return ConstantNumeric(action11.add_sound((args[0].value, volume), pos), pos)


@builtin
def builtin_import_sound(name, args, pos):
    if len(args) not in (2, 3):
        raise generic.ScriptError(name + "() must have 2 or 3 parameters", pos)
    grfid = parse_string_to_dword(args[0].reduce())
    sound_num = args[1].reduce_constant().value
    volume = args[2].reduce_constant().value if len(args) >= 3 else 100
    generic.check_range(volume, 0, 100, "sound volume", pos)
    from nml.actions import action11

    return ConstantNumeric(action11.add_sound((grfid, sound_num, volume), pos), pos)


@builtin
def builtin_relative_coord(name, args, pos):
    """
    relative_coord(x, y) builtin function.

    @return Coordinates in 0xYYXX format.
    """
    if len(args) != 2:
        raise generic.ScriptError(name + "() must have x and y coordinates as parameters", pos)

    if isinstance(args[0], ConstantNumeric):
        generic.check_range(args[0].value, 0, 255, "Argument of '{}'".format(name), args[0].pos)
    if isinstance(args[1], ConstantNumeric):
        generic.check_range(args[1].value, 0, 255, "Argument of '{}'".format(name), args[1].pos)

    x_coord = nmlop.AND(args[0], 0xFF)
    y_coord = nmlop.AND(args[1], 0xFF)
    # Shift Y to its position.
    y_coord = nmlop.SHIFT_LEFT(y_coord, 8)

    return nmlop.OR(x_coord, y_coord, pos)


@builtin
def builtin_num_corners_raised(name, args, pos):
    """
    num_corners_raised(slope) builtin function.
    slope is a 5-bit value

    @return Number of raised corners in a slope (4 for steep slopes)
    """
    if len(args) != 1:
        raise generic.ScriptError(name + "() must have 1 parameter", pos)

    slope = args[0]
    # The returned value is ((slope x 0x8421) & 0x11111) % 0xF
    # Explanation in steps: (numbers in binary)
    # - Masking constrains the slope to 5 bits, just to be sure (a|bcde)
    # - Multiplication creates 4 copies of those bits (abcd|eabc|deab|cdea|bcde)
    # - And-masking leaves only the lowest bit in each nibble (000d|000c|000b|000a|000e)
    # - The modulus operation adds one to the output for each set bit
    # - We now have the count of bits in the slope, which is wat we want. yay!
    slope = nmlop.AND(slope, 0x1F, pos)
    slope = nmlop.MUL(slope, 0x8421)
    slope = nmlop.AND(slope, 0x11111)
    return nmlop.MOD(slope, 0xF)


@builtin
def builtin_slope_to_sprite_offset(name, args, pos):
    """
    builtin function slope_to_sprite_offset(slope)

    @return sprite offset to use
    """
    if len(args) != 1:
        raise generic.ScriptError(name + "() must have 1 parameter", pos)

    if isinstance(args[0], ConstantNumeric):
        generic.check_range(args[0].value, 0, 15, "Argument of '{}'".format(name), args[0].pos)

    # step 1: ((slope >= 0) & (slope <= 14)) * slope
    # This handles all non-steep slopes
    expr = nmlop.AND(nmlop.CMP_LE(args[0], 14, pos), nmlop.CMP_GE(args[0], 0, pos))
    expr = nmlop.MUL(expr, args[0])
    # Now handle the steep slopes separately
    # So add (slope == SLOPE_XX) * offset_of_SLOPE_XX for each steep slope
    steep_slopes = [(23, 16), (27, 17), (29, 15), (30, 18)]
    for slope, offset in steep_slopes:
        to_add = nmlop.MUL(nmlop.CMP_EQ(args[0], slope, pos), offset)
        expr = nmlop.ADD(expr, to_add)
    return expr


@builtin
def builtin_palette_1cc(name, args, pos):
    """
    palette_1cc(colour) builtin function.

    @return Recolour sprite to use
    """
    if len(args) != 1:
        raise generic.ScriptError(name + "() must have 1 parameter", pos)

    if isinstance(args[0], ConstantNumeric):
        generic.check_range(args[0].value, 0, 15, "Argument of '{}'".format(name), args[0].pos)

    return nmlop.ADD(args[0], 775, pos)


@builtin
def builtin_palette_2cc(name, args, pos):
    """
    palette_2cc(colour1, colour2) builtin function.

    @return Recolour sprite to use
    """
    if len(args) != 2:
        raise generic.ScriptError(name + "() must have 2 parameters", pos)

    for i in range(0, 2):
        if isinstance(args[i], ConstantNumeric):
            generic.check_range(args[i].value, 0, 15, "Argument of '{}'".format(name), args[i].pos)

    col2 = nmlop.MUL(args[1], 16, pos)
    col12 = nmlop.ADD(col2, args[0])
    # Base sprite is not a constant
    base = global_constants.patch_variable("base_sprite_2cc", global_constants.patch_variables["base_sprite_2cc"], pos)

    return nmlop.ADD(col12, base)


@builtin
def builtin_vehicle_curv_info(name, args, pos):
    """
    vehicle_curv_info(prev_cur, cur_next) builtin function

    @return Value to use with vehicle var curv_info
    """
    if len(args) != 2:
        raise generic.ScriptError(name + "() must have 2 parameters", pos)

    for arg in args:
        if isinstance(arg, ConstantNumeric):
            generic.check_range(arg.value, -2, 2, "Argument of '{}'".format(name), arg.pos)

    args = [nmlop.AND(arg, 0xF, pos) for arg in args]
    cur_next = nmlop.SHIFT_LEFT(args[1], 8)
    return nmlop.OR(args[0], cur_next)


@builtin
def builtin_format_string(name, args, pos):
    """
    format_string(format, ... args ..) builtin function

    @return Formatted string
    """
    if len(args) < 1:
        raise generic.ScriptError(name + "() must have at least one parameter", pos)

    format = args[0].reduce()
    if not isinstance(format, StringLiteral):
        raise generic.ScriptError(name + "() parameter 1 'format' must be a literal string", format.pos)

    # Validate other args
    format_args = []
    for i, arg in enumerate(args[1:]):
        arg = arg.reduce()
        if not isinstance(arg, (StringLiteral, ConstantFloat, ConstantNumeric)):
            raise generic.ScriptError(
                name + "() parameter {:d} is not a constant number of literal string".format(i + 1), arg.pos
            )
        format_args.append(arg.value)

    try:
        result = format.value % tuple(format_args)
        return StringLiteral(result, pos)
    except Exception as ex:
        raise generic.ScriptError("Invalid combination of format / arguments for {}: {}".format(name, str(ex)), pos)


# }



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

import os
import sys
import time

# Enable VT100 sequences on windows consoles
if os.name == "nt":

    class Break(Exception):
        pass

    try:
        from ctypes import byref, windll
        from ctypes.wintypes import DWORD, HANDLE

        kernel32 = windll.kernel32
        h = kernel32.GetStdHandle(-11)  # stdout
        if h is None or h == HANDLE(-1):
            raise Break()
        FILE_TYPE_CHAR = 0x0002
        if (kernel32.GetFileType(h) & 3) != FILE_TYPE_CHAR:
            raise Break()
        mode = DWORD()
        kernel32.GetConsoleMode(h, byref(mode))
        ENABLE_VIRTUAL_TERMINAL_PROCESSING = 0x0004
        if (mode.value & ENABLE_VIRTUAL_TERMINAL_PROCESSING) == 0:
            kernel32.SetConsoleMode(h, mode.value | ENABLE_VIRTUAL_TERMINAL_PROCESSING)
    except Break:
        # Early exit
        pass


def truncate_int32(value):
    """
    Truncate the given value so it can be stored in exactly 4 bytes. The sign
    will be kept. Too high or too low values will be cut off, not clamped to
    the valid range.

    @param value: The value to truncate.
    @type value: C{int}

    @return: The truncated value.
    @rtype: C{int}.
    """
    # source: http://www.tiac.net/~sw/2010/02/PureSalsa20/index.html
    return int((value & 0x7FFFFFFF) | -(value & 0x80000000))


def check_range(value, min_value, max_value, name, pos):
    """
    Check if a value is within a certain range and raise an error if it's not.

    @param value: The value to check.
    @type value: C{int}

    @param min_value: Minimum valid value.
    @type min_value: C{int}

    @param max_value: Maximum valid value.
    @type max_value: C{int}

    @param name: Name of the variable that is being tested.
    @type name: C{str}

    @param pos: Position information from the variable being tested.
    @type pos: L{Position}
    """
    if not min_value <= value <= max_value:
        raise RangeError(value, min_value, max_value, name, pos)


def greatest_common_divisor(a, b):
    """
    Get the greatest common divisor of two numbers

    @param a: First number.
    @type  a: C{int}

    @param b: Second number.
    @type  b: C{int}

    @return: Greatest common divisor.
    @rtype:  C{int}
    """
    while b != 0:
        t = b
        b = a % b
        a = t
    return a


def reverse_lookup(dic, val):
    """
    Perform reverse lookup of any key that has the provided value.

    @param dic: Dictionary to perform reverse lookup on.
    @type  dic: C{dict}

    @param val: Value being searched.
    @type  val: an existing value.

    @return: A key such that C{dict[key] == val}.
    @rtype:  Type of the matching key.
    """
    for k, v in dic.items():
        if v == val:
            return k
    raise AssertionError("Value not found in the dictionary.")


def build_position(poslist):
    """
    Construct a L{Position} object that takes the other positions as include list.

    @param poslist: Sequence of positions to report. First entry is the innermost position,
                    last entry is the nml statement that started it all.
    @type  poslist: C{list} of L{Position}

    @return: Position to attach to an error.
    @rtype:  L{Position}
    """
    if poslist is None or len(poslist) == 0:
        return None
    if len(poslist) == 1:
        return poslist[0]
    pos = poslist[-1]
    pos.includes = pos.includes + poslist[:-1]
    return pos


class Position:
    """
    Base class representing a position in a file.

    @ivar filename: Name of the file.
    @type filename: C{str}

    @ivar includes: List of file includes
    @type includes: C{list} of L{Position}
    """

    def __init__(self, filename, includes):
        self.filename = filename
        self.includes = includes


class LinePosition(Position):
    """
    Line in a file.

    @ivar line_start: Line number (starting with 1) where the position starts.
    @type line_start: C{int}
    """

    def __init__(self, filename, line_start, includes=None):
        Position.__init__(self, filename, includes or [])
        self.line_start = line_start

    def __str__(self):
        return '"{}", line {:d}'.format(self.filename, self.line_start)


class PixelPosition(Position):
    """
    Position of a pixel (in a file with graphics).

    @ivar xpos: Horizontal position of the pixel.
    @type xpos: C{int}

    @ivar ypos: Vertical position of the pixel.
    @type ypos: C{int}
    """

    def __init__(self, filename, xpos, ypos):
        Position.__init__(self, filename, [])
        self.xpos = xpos
        self.ypos = ypos

    def __str__(self):
        return '"{}" at [x: {:d}, y: {:d}]'.format(self.filename, self.xpos, self.ypos)


class ImageFilePosition(Position):
    """
    Generic (not position-dependant) error with an image file
    """

    def __init__(self, filename, pos=None):
        poslist = []
        if pos is not None:
            poslist.append(pos)
        Position.__init__(self, filename, poslist)

    def __str__(self):
        return 'Image file "{}"'.format(self.filename)


class LanguageFilePosition(Position):
    """
    Generic (not position-dependant) error with a language file.
    """

    def __init__(self, filename):
        Position.__init__(self, filename, [])

    def __str__(self):
        return 'Language file "{}"'.format(self.filename)


class ScriptError(Exception):
    def __init__(self, value, pos=None):
        self.value = value
        self.pos = pos

    def __str__(self):
        if self.pos is None:
            return self.value
        else:
            ret = str(self.pos) + ": " + self.value
            for inc in reversed(self.pos.includes):
                ret += "\nIncluded from: " + str(inc)
            return ret


class ConstError(ScriptError):
    """
    Error to denote a compile-time integer constant was expected but not found.
    """

    def __init__(self, pos=None):
        ScriptError.__init__(self, "Expected a compile-time integer constant", pos)


class RangeError(ScriptError):
    def __init__(self, value, min_value, max_value, name, pos=None):
        ScriptError.__init__(
            self, name + " out of range " + str(min_value) + ".." + str(max_value) + ", encountered " + str(value), pos
        )


class ProcCallSyntaxError(ScriptError):
    def __init__(self, name, pos=None):
        ScriptError.__init__(self, "Missing '()' after '{}'.".format(name), pos)


class ImageError(ScriptError):
    def __init__(self, value, filename, pos=None):
        ScriptError.__init__(self, value, ImageFilePosition(filename, pos))


class OnlyOnceError(ScriptError):
    """
    An error denoting two elements in a single grf were found, where only one is allowed.
    """

    def __init__(self, typestr, pos=None):
        """
        @param typestr: Description of the type of element encountered.
        @type  typestr: C{str}

        @param pos: Position of the error, if provided.
        @type  pos: C{None} or L{Position}
        """
        ScriptError.__init__(self, "A grf may contain only one {}.".format(typestr), pos)


class OnlyOnce:
    """
    Class to enforce that certain objects / constructs appear only once.
    """

    seen = {}

    @classmethod
    def enforce(cls, obj, typestr):
        """
        If this method is called more than once for an object of the exact same
        class, an OnlyOnceError is raised.
        """
        objtype = obj.__class__
        if objtype in cls.seen:
            raise OnlyOnceError(typestr, obj.pos)
        cls.seen[objtype] = None

    @classmethod
    def clear(cls):
        cls.seen = {}


class Warning:
    GENERIC = 0
    DEPRECATION = 1
    OPTIMISATION = 2
    disabled = None


VERBOSITY_WARNING = 1  # Verbosity level for warnings
VERBOSITY_INFO = 2  # Verbosity level for info messages
VERBOSITY_PROGRESS = 3  # Verbosity level for progress feedback
VERBOSITY_TIMING = 4  # Verbosity level for timing information

VERBOSITY_MAX = 4  # Maximum verbosity level

"""
Verbosity level for console output.
"""
verbosity_level = VERBOSITY_PROGRESS


def set_verbosity(level):
    global verbosity_level
    verbosity_level = level


def print_eol(msg):
    """
    Clear current line and print message without linefeed.
    """
    if not os.isatty(sys.stdout.fileno()):
        return

    print("\r" + msg + "\033[K", end="")


"""
Current progress message.
"""
progress_message = None

"""
Timestamp when the current processing step started.
"""
progress_start_time = None

"""
Timestamp of the last incremental progress update.
"""
progress_update_time = None


def hide_progress():
    if progress_message is not None:
        print_eol("")


def show_progress():
    if progress_message is not None:
        print_eol(progress_message)


def clear_progress():
    global progress_message
    global progress_start_time
    global progress_update_time
    hide_progress()

    if (progress_message is not None) and (verbosity_level >= VERBOSITY_TIMING):
        print("{} {:.1f} s".format(progress_message, time.process_time() - progress_start_time))

    progress_message = None
    progress_start_time = None
    progress_update_time = None


def print_progress(msg, incremental=False):
    """
    Output progress information to the user.

    @param msg: Progress message.
    @type  msg: C{str}

    @param incremental: True if this message is updated incrementally (that is, very often).
    @type  incremental: C{bool}
    """
    if verbosity_level < VERBOSITY_PROGRESS:
        return

    global progress_message
    global progress_start_time
    global progress_update_time

    if (not incremental) and (progress_message is not None):
        clear_progress()

    progress_message = msg

    if incremental:
        t = time.process_time()
        if (progress_update_time is not None) and (t - progress_update_time < 1):
            return
        progress_update_time = t
    else:
        progress_start_time = time.process_time()

    print_eol(msg)


def print_info(msg):
    """
    Output a pure informational message to the user.
    """
    if verbosity_level < VERBOSITY_INFO:
        return

    hide_progress()
    print(" nmlc info: " + msg)
    show_progress()


def print_warning(type, msg, pos=None):
    """
    Output a warning message to the user.
    """
    if verbosity_level < VERBOSITY_WARNING:
        return
    if Warning.disabled and type in Warning.disabled:
        return
    if pos:
        msg = str(pos) + ": " + msg

    msg = " nmlc warning: " + msg

    if sys.stderr.isatty():
        msg = "\033[93m" + msg + "\033[0m"

    hide_progress()
    print(msg, file=sys.stderr)
    show_progress()


def print_error(msg, pos=None):
    """
    Output an error message to the user.
    """
    if pos:
        msg = str(pos) + ": " + msg
    else:
        clear_progress()

    msg = " nmlc ERROR: " + msg

    if sys.stderr.isatty():
        msg = "\033[91m" + msg + "\033[0m"

    hide_progress()
    print(msg, file=sys.stderr)
    show_progress()


def print_dbg(indent, *args):
    """
    Output debug text.

    @param indent: Indentation to add to the output.
    @type  indent: C{int}

    @param args: Arguments to print. An additional space is printed between them.
    @type  args: C{Tuple} of C{str}
    """
    hide_progress()
    print(indent * " " + " ".join(str(arg) for arg in args))
    show_progress()


"""
Paths already resolved to correct paths on the system.
The key is the path as specified in the sources. The value is the validated path on the system.
"""
_paths = {}


def find_file(filepath):
    """
    Verify whether L{filepath} exists. If not, try to find a similar one with a
    different case.

    @param filepath: Path to the file to open.
    @type  filepath: C{str}

    @return: Path name to a file that exists at the file system.
    @rtype:  C{str}
    """
    # Split the filepath in components (directory parts and the filename).
    drive, filepath = os.path.splitdrive(os.path.normpath(filepath))
    # 'splitdrive' above does not remove the leading / of absolute Unix paths.
    # The 'split' below splits on os.sep, which means that loop below fails for "/".
    # To prevent that, handle the leading os.sep separately.
    if filepath.startswith(os.sep):
        drive = drive + os.sep
        filepath = filepath[len(os.sep) :]

    components = []  # Path stored in reverse order (filename at index[0])
    while filepath != "":
        filepath, filepart = os.path.split(filepath)
        components.append(filepart)

    # Re-build the absolute path.
    path = drive
    if path == "":
        path = os.getcwd()
    while len(components) > 0:
        comp = components.pop()
        childpath = os.path.join(path, comp)
        if childpath in _paths:
            path = _paths[childpath]
            continue

        if os.access(path, os.R_OK):
            # Path is readable, compare provided path with the file system.
            entries = os.listdir(path) + [os.curdir, os.pardir]
            lcomp = comp.lower()
            matches = [entry for entry in entries if lcomp == entry.lower()]

            if len(matches) == 0:
                raise ScriptError(
                    'Path "{}" does not exist (even after case conversions)'.format(os.path.join(path, comp))
                )
            elif len(matches) > 1:
                raise ScriptError(
                    'Path "{}" is not unique (case conversion gave {:d} solutions)'.format(
                        os.path.join(path, comp), len(matches)
                    )
                )

            if matches[0] != comp:
                given_path = os.path.join(path, comp)
                real_path = os.path.join(path, matches[0])
                msg = (
                    'Path "{}" at the file system does not match path "{}" given in the input'
                    " (case mismatch in the last component)"
                ).format(real_path, given_path)
                print_warning(Warning.GENERIC, msg)
        elif os.access(path, os.X_OK):
            # Path is only accessible, cannot inspect the file system.
            matches = [comp]
        else:
            raise ScriptError('Path "{}" does not exist or is not accessible'.format(path))

        path = os.path.join(path, matches[0])
        if len(components) > 0:
            _paths[childpath] = path

    return path


cache_root_dir = ".nmlcache"


def set_cache_root_dir(dir):
    global cache_root_dir
    cache_root_dir = None if dir is None else os.path.abspath(dir)


def _cache_file_path(sources, extension):
    """
    Compose a filename for a cache file.

    @param sources: List of source files, the cache file depends on / belongs to.
    @type  sources: C{list} or C{tuple} of C{str} or similar.

    @param extension: File extension for the cache file including leading ".".
    @type  extension: C{str}

    @return: Filename for cache file.
    @rtype:  C{str}
    """
    result = ""

    for part in sources:
        if part is not None:
            path, name = os.path.split(part)

            if len(result) == 0:
                # Make sure that the path does not leave the cache dir
                path = os.path.normpath(path).replace(os.path.pardir, "__")
                path = os.path.join(cache_root_dir, path)
                result = os.path.join(path, name)
            else:
                # In case of multiple source files, ignore the path component for all but the first
                result += "_" + name

    return result + extension


def open_cache_file(sources, extension, mode):
    if cache_root_dir is None:
        raise FileNotFoundError("No cache directory")

    if not any(sources):
        raise FileNotFoundError("Can't create cache file with no sources")

    path = _cache_file_path(sources, extension)

    try:
        if "w" in mode:
            os.makedirs(os.path.dirname(path), exist_ok=True)
        return open(path, mode)
    except OSError:
        if "w" in mode:
            print_warning(
                Warning.GENERIC,
                "Can't create cache file {}. Check permissions, or use --cache-dir or --no-cache.".format(path),
            )
        raise



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import generic

from .base_expression import Expression


class PatchVariable(Expression):
    """
    Class for reading so-called 'patch variables' via a special ActionD

    @ivar num: Variable number to read
    @type num: C{int}
    """

    def __init__(self, num, pos=None):
        Expression.__init__(self, pos)
        self.num = num

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "PatchVariable:", self.num)

    def __str__(self):
        return "PatchVariable({:d})".format(self.num)

    def reduce(self, id_dicts=None, unknown_id_fatal=True):
        return self

    def supported_by_action2(self, raise_error):
        if raise_error:
            raise generic.ScriptError("Reading patch variables is not supported in a switch-block.", self.pos)
        return False

    def supported_by_actionD(self, raise_error):
        return True



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import grfstrings
from nml.actions import base_action


class Action8(base_action.BaseAction):
    def __init__(self, grfid, name, description):
        self.grfid = grfid
        self.name = name
        self.description = description

    def write(self, file):
        name = grfstrings.get_translation(self.name)
        desc = grfstrings.get_translation(self.description)
        size = 6 + grfstrings.get_string_size(name) + grfstrings.get_string_size(desc)
        file.start_sprite(size)
        file.print_bytex(8)
        # Write the grf version
        file.print_bytex(8)
        file.print_string(self.grfid.value, False, True)
        file.print_string(name)
        file.print_string(desc)
        file.end_sprite()

    def skip_action7(self):
        return False



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

"""
Action 11 support classes (sounds).
"""
import os

from nml import generic
from nml.actions import action0, base_action


class Action11(base_action.BaseAction):
    def __init__(self, num_sounds):
        self.num_sounds = num_sounds

    def write(self, file):
        file.start_sprite(3)
        file.print_bytex(0x11)
        file.print_word(self.num_sounds)
        file.end_sprite()


class LoadBinaryFile(base_action.BaseAction):
    """
    <sprite-number> * <length> FF <name-len> <name> 00 <data>
    """

    def __init__(self, fname, pos):
        self.fname = fname
        self.pos = pos
        self.last = False

    def prepare_output(self, sprite_num):
        if not os.access(self.fname, os.R_OK):
            raise generic.ScriptError("Sound file '{}' does not exist.".format(self.fname), self.pos)
        size = os.path.getsize(self.fname)
        if size == 0:
            raise generic.ScriptError(
                "Expected sound file '{}' to have a non-zero length.".format(self.fname), self.pos
            )

    def write(self, file):
        file.print_named_filedata(self.fname)
        if self.last:
            file.newline()


class ImportSound(base_action.BaseAction):
    """
    Import a sound from another grf::

        <sprite-number> * <length> FE 00 <grfid> <number>

    @ivar grfid: ID of the other grf.
    @type grfid: C{int}

    @ivar number: Sound number to load.
    @type number: C{int}

    @ivar pos: Position information
    @type pos: L{Position}
    """

    def __init__(self, grfid, number, pos):
        self.grfid = grfid
        self.number = number
        self.pos = pos
        self.last = False

    def write(self, file):
        file.start_sprite(8)
        file.print_bytex(0xFE)
        file.print_bytex(0)
        file.print_dwordx(self.grfid)
        file.print_wordx(self.number)
        file.end_sprite()
        if self.last:
            file.newline()


registered_sounds = {}
SOUND_OFFSET = 73  # No of original sounds
NUM_ANIMATION_SOUNDS = 0x80 - SOUND_OFFSET  # Number of custom sound ids, which can be returned by animation callbacks.


def print_stats():
    """
    Print statistics about used ids.
    """
    if len(registered_sounds) > 0:
        # Currently NML does not optimise the order of sound effects. So we assume NUM_ANIMATION_SOUNDS as the maximum.
        generic.print_info("Sound effects: {}/{}".format(len(registered_sounds), NUM_ANIMATION_SOUNDS))


def add_sound(args, pos):
    if args not in registered_sounds:
        registered_sounds[args] = (len(registered_sounds), pos)
    return registered_sounds[args][0] + SOUND_OFFSET


def get_sound_actions():
    """
    Get a list of actions that actually includes all sounds in the output file.
    """
    if not registered_sounds:
        return []

    action_list = []
    action_list.append(Action11(len(registered_sounds)))
    volume_list = []

    sound_data = [(sound_id, args, pos) for args, (sound_id, pos) in registered_sounds.items()]
    # Sort on first item, i.e. sound ID
    for i, args, pos in sorted(sound_data):
        if len(args) == 3:
            action_list.append(ImportSound(args[0], args[1], pos))
        else:
            action_list.append(LoadBinaryFile(args[0], pos))
        if args[-1] != 100:
            volume_list.append((i + SOUND_OFFSET, int(args[-1] * 128 / 100)))

    if volume_list:
        action_list.extend(action0.get_volume_actions(volume_list))

    return action_list



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import generic
from nml.actions import base_action, real_sprite


class Action5(base_action.BaseAction):
    def __init__(self, type, num_sprites, offset):
        self.type = type
        self.num_sprites = num_sprites
        self.offset = offset

    def prepare_output(self, sprite_num):
        if self.offset is not None:
            self.type |= 0x80

    def write(self, file):
        # <Sprite-number> * <Length> 05 <type> <num-sprites> [<offset>]
        size = 5 if self.offset is None else 8
        file.start_sprite(size)
        file.print_bytex(0x05)
        file.print_bytex(self.type)
        file.print_bytex(0xFF)
        file.print_word(self.num_sprites)
        if self.offset is not None:
            file.print_bytex(0xFF)
            file.print_word(self.offset)
        file.newline()
        file.end_sprite()

    def skip_action7(self):
        # skipping with Action7 should work, according to the Action7/9 specs
        # However, skipping invalid (OpenTTD-only) Action5s in TTDP can only be done using Action9, else an error occurs
        # To be on the safe side, don't allow skipping with Action7 at all
        return False


class Action5BlockType:
    FIXED = (0,)  # fixed number of sprites
    ANY = (1,)  # any number of sprites
    OFFSET = (2,)  # flexible number of sprites, offset may be set


action5_table = {
    "PRE_SIGNAL": (0x04, 48, Action5BlockType.OFFSET),  # deprecated, use "SIGNALS" in all cases
    "PRE_SIGNAL_SEMAPHORE": (0x04, 112, Action5BlockType.OFFSET),  # deprecated, use "SIGNALS" in all cases
    "PRE_SIGNAL_SEMAPHORE_PBS": (0x04, 240, Action5BlockType.OFFSET),  # deprecated, use "SIGNALS" in all cases
    "SIGNALS": (0x04, 240, Action5BlockType.OFFSET),
    "CATENARY": (0x05, 48, Action5BlockType.OFFSET),
    "FOUNDATIONS_SLOPES": (0x06, 74, Action5BlockType.FIXED),
    "FOUNDATIONS_SLOPES_HALFTILES": (0x06, 90, Action5BlockType.OFFSET),
    "TTDP_GUI_25": (0x07, 73, Action5BlockType.FIXED),
    "TTDP_GUI": (0x07, 93, Action5BlockType.FIXED),
    "CANALS": (0x08, 65, Action5BlockType.OFFSET),
    "ONE_WAY_ROAD": (0x09, 18, Action5BlockType.OFFSET),
    "COLOURMAP_2CC": (0x0A, 256, Action5BlockType.OFFSET),
    "TRAMWAY": (0x0B, 119, Action5BlockType.OFFSET),
    "SNOWY_TEMPERATE_TREES": (0x0C, 133, Action5BlockType.FIXED),
    "COAST_TILES": (0x0D, 16, Action5BlockType.FIXED),
    "COAST_TILES_BASEGFX": (0x0D, 10, Action5BlockType.FIXED),
    "COAST_TILES_DIAGONAL": (0x0D, 18, Action5BlockType.FIXED),
    "NEW_SIGNALS": (0x0E, 0, Action5BlockType.ANY),
    "SLOPED_RAILS": (0x0F, 12, Action5BlockType.OFFSET),
    "AIRPORTS": (0x10, 15, Action5BlockType.OFFSET),
    "ROAD_STOPS": (0x11, 8, Action5BlockType.OFFSET),
    "AQUEDUCTS": (0x12, 8, Action5BlockType.OFFSET),
    "AUTORAIL": (0x13, 55, Action5BlockType.OFFSET),
    "FLAGS": (0x14, 36, Action5BlockType.OFFSET),
    "OTTD_GUI": (0x15, 192, Action5BlockType.OFFSET),
    "AIRPORT_PREVIEW": (0x16, 9, Action5BlockType.OFFSET),
    "RAILTYPE_TUNNELS": (0x17, 16, Action5BlockType.OFFSET),
    "OTTD_RECOLOUR": (0x18, 1, Action5BlockType.OFFSET),
    "ROAD_WAYPOINTS": (0x19, 4, Action5BlockType.OFFSET),
    "OVERLAY_ROCKS": (0x1A, 95, Action5BlockType.OFFSET),
    "BRIDGE_DECKS": (0x1B, 24, Action5BlockType.OFFSET),
}


def parse_action5(replaces):
    real_sprite_list = real_sprite.parse_sprite_data(replaces)
    num_sprites = len(real_sprite_list)

    if replaces.type.value not in action5_table:
        raise generic.ScriptError(replaces.type.value + " is not a valid sprite replacement type", replaces.type.pos)
    type_id, num_required, block_type = action5_table[replaces.type.value]
    offset = None

    if block_type == Action5BlockType.FIXED:
        if num_sprites < num_required:
            msg = "Invalid sprite count for sprite replacement type '{}', expected {:d}, got {:d}"
            msg = msg.format(replaces.type, num_required, num_sprites)
            raise generic.ScriptError(msg, replaces.pos)

        elif num_sprites > num_required:
            msg = (
                "Too many sprites specified for sprite replacement type '{}',"
                " expected {:d}, got {:d}, extra sprites may be ignored"
            ).format(replaces.type, num_required, num_sprites)
            generic.print_warning(generic.Warning.GENERIC, msg, replaces.pos)

        if replaces.offset != 0:
            msg = "replacenew parameter 'offset' must be zero for sprite replacement type '{}'".format(replaces.type)
            raise generic.ScriptError(msg, replaces.pos)

    elif block_type == Action5BlockType.ANY:
        if replaces.offset != 0:
            msg = "replacenew parameter 'offset' must be zero for sprite replacement type '{}'".format(replaces.type)
            raise generic.ScriptError(msg, replaces.pos)

    elif block_type == Action5BlockType.OFFSET:
        if num_sprites + replaces.offset > num_required:
            msg = "Exceeding the limit of {:d} sprites for sprite replacement type '{}', extra sprites may be ignored"
            msg = msg.format(num_required, replaces.type)
            generic.print_warning(generic.Warning.GENERIC, msg, replaces.pos)

        if replaces.offset != 0 or num_sprites != num_required:
            offset = replaces.offset
    else:
        assert 0

    return [Action5(type_id, num_sprites, offset)] + real_sprite_list



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import expression, free_number_list, generic, nmlop
from nml.actions import action6, action10, actionD, base_action

free_labels = free_number_list.FreeNumberList(
    list(range(0xFF, 0x0F, -1)),
    "No label available to use for large if-blocks and loops.",
    "No unique label available to use for large if-blocks and loops.",
)


def print_stats():
    """
    Print statistics about used ids.
    """
    if free_labels.stats[0] > 0:
        generic.print_info(
            "Concurrent Action10 labels: {}/{} ({})".format(
                free_labels.stats[0], free_labels.total_amount, str(free_labels.stats[1])
            )
        )


class SkipAction(base_action.BaseAction):
    def __init__(self, action_type, var, varsize, condtype, value, label):
        self.action_type = action_type
        self.label = label
        self.var = var
        self.varsize = varsize
        self.condtype = condtype
        self.value = value
        self.label = label
        if self.condtype[0] == 0 or self.condtype[0] == 1:
            assert self.varsize == 1

    def write(self, file):
        size = 5 + self.varsize
        file.start_sprite(size)
        file.print_bytex(self.action_type)
        file.print_bytex(self.var)
        file.print_bytex(self.varsize)
        file.print_bytex(self.condtype[0], self.condtype[1])
        if self.varsize == 8:
            # grfid + mask
            file.print_dwordx(self.value & 0xFFFFFFFF)
            file.print_dwordx(self.value >> 32)
        else:
            file.print_varx(self.value, self.varsize)
        file.print_bytex(self.label)
        file.newline()
        file.end_sprite()

    def skip_action7(self):
        return self.action_type == 7

    def skip_action9(self):
        return self.action_type == 9 or self.label == 0


class UnconditionalSkipAction(SkipAction):
    def __init__(self, action_type, label):
        SkipAction.__init__(self, action_type, 0x9A, 1, (0, r"\71"), 0, label)


def op_to_cond_op(op):
    # The operators are reversed as we want to skip if the expression is true
    # while the nml-syntax wants to execute the block if the expression is true
    if op == nmlop.CMP_NEQ:
        return (2, r"\7=")
    if op == nmlop.CMP_EQ:
        return (3, r"\7!")
    if op == nmlop.CMP_GE:
        return (4, r"\7<")
    if op == nmlop.CMP_LE:
        return (5, r"\7>")
    # Not reached
    raise ValueError("Unexpected operator '{}'".format(op.token))


def parse_conditional(expr):
    """
    Parse an expression and return enough information to use
    that expression as a conditional statement.
    Return value is a tuple with the following elements:
    - Parameter number (as integer) to use in comparison or None for unconditional skip
    - List of actions needed to set the given parameter to the correct value
    - The type of comparison to be done
    - The value to compare against (as integer)
    - The size of the value (as integer)
    """
    if expr is None:
        return (None, [], (2, r"\7="), 0, 4)
    if isinstance(expr, expression.BinOp):
        if expr.op == nmlop.HASBIT or expr.op == nmlop.NOTHASBIT:
            if isinstance(expr.expr1, expression.Parameter) and isinstance(expr.expr1.num, expression.ConstantNumeric):
                param = expr.expr1.num.value
                actions = []
            else:
                param, actions = actionD.get_tmp_parameter(expr.expr1)
            if isinstance(expr.expr2, expression.ConstantNumeric):
                bit_num = expr.expr2.value
            else:
                if isinstance(expr.expr2, expression.Parameter) and isinstance(
                    expr.expr2.num, expression.ConstantNumeric
                ):
                    param = expr.expr2.num.value
                else:
                    param, tmp_action_list = actionD.get_tmp_parameter(expr.expr2)
                    actions.extend(tmp_action_list)
                act6 = action6.Action6()
                act6.modify_bytes(param, 1, 4)
                actions.append(act6)
                bit_num = 0
            comp_type = (1, r"\70") if expr.op == nmlop.HASBIT else (0, r"\71")
            return (param, actions, comp_type, bit_num, 1)
        elif expr.op in (nmlop.CMP_EQ, nmlop.CMP_NEQ, nmlop.CMP_LE, nmlop.CMP_GE) and isinstance(
            expr.expr2, expression.ConstantNumeric
        ):
            if isinstance(expr.expr1, expression.Parameter) and isinstance(expr.expr1.num, expression.ConstantNumeric):
                param = expr.expr1.num.value
                actions = []
            else:
                param, actions = actionD.get_tmp_parameter(expr.expr1)
            op = op_to_cond_op(expr.op)
            return (param, actions, op, expr.expr2.value, 4)

    if isinstance(expr, expression.Boolean):
        expr = expr.expr

    if isinstance(expr, expression.Not):
        param, actions = actionD.get_tmp_parameter(expr.expr)
        return (param, actions, (3, r"\7!"), 0, 4)

    param, actions = actionD.get_tmp_parameter(expr)
    return (param, actions, (2, r"\7="), 0, 4)


def cond_skip_actions(action_list, param, condtype, value, value_size, pos):
    if len(action_list) == 0:
        return []
    actions = []
    start, length = 0, 0
    # Whether to allow not-skipping, using action7 or using action9
    skip_opts = (True, True, True)
    # Add a sentinel value to the list to avoid code duplication
    for action in action_list + [None]:
        assert any(skip_opts)
        if action is not None:
            # Options allowed by the next action
            act_opts = (not action.skip_needed(), action.skip_action7(), action.skip_action9())
        else:
            # There are no further actions, so be sure to finish the current block
            act_opts = (False, False, False)
        # Options still allowed, when including this action in the previous block
        new_opts = tuple(all(tpl) for tpl in zip(skip_opts, act_opts))
        # End this block in two cases:
        # - There are no options to skip this action and the preceding block in one go
        # - The existing block (with length > 0) needed no skipping, but this action does
        if any(new_opts) and not (length > 0 and skip_opts[0] and not new_opts[0]):
            # Append this action to the existing block
            length += 1
            skip_opts = new_opts
            continue

        # We need to create a new block
        if skip_opts[0]:
            # We can just choose to not skip the preceding actions without harm
            actions.extend(action_list[start : start + length])
        else:
            action_type = 7 if skip_opts[1] else 9
            if length < 0x10:
                # Lengths under 0x10 are handled without labels, to avoid excessive label usage
                target = length
                label = None
            else:
                target = free_labels.pop(pos)
                label = action10.Action10(target)
            actions.append(SkipAction(action_type, param, value_size, condtype, value, target))
            actions.extend(action_list[start : start + length])
            if label is not None:
                actions.append(label)

        start = start + length
        length = 1
        skip_opts = act_opts
    assert start == len(action_list)

    return actions


recursive_cond_blocks = 0


def parse_conditional_block(cond_list):
    global recursive_cond_blocks
    recursive_cond_blocks += 1
    if recursive_cond_blocks == 1:
        # We only save a single state (at toplevel nml-blocks) because
        # we don't know at the start of the block how many labels we need.
        # Getting the same label for a block that was already used in a
        # sub-block would be very bad, since the action7/9 would skip
        # to the action10 of the sub-block.
        free_labels.save()

    blocks = []
    for cond in cond_list.statements:
        if isinstance(cond.expr, expression.ConstantNumeric):
            if cond.expr.value == 0:
                continue
            else:
                blocks.append({"expr": None, "statements": cond.statements})
                break
        blocks.append({"expr": cond.expr, "statements": cond.statements})
    if blocks:
        blocks[-1]["last_block"] = True

    if len(blocks) == 1 and blocks[0]["expr"] is None:
        action_list = []
        for stmt in blocks[0]["statements"]:
            action_list.extend(stmt.get_action_list())
        return action_list

    action6.free_parameters.save()

    if len(blocks) > 1:
        # the skip all parameter is used to skip all blocks after one
        # of the conditionals was true. We can't always skip directly
        # to the end of the blocks since action7/action9 can't always
        # be mixed
        param_skip_all, action_list = actionD.get_tmp_parameter(expression.ConstantNumeric(0xFFFFFFFF))
    else:
        action_list = []

    # use parse_conditional here, we also need to know if all generated
    # actions (like action6) can be skipped safely
    for block in blocks:
        (
            block["param_dst"],
            block["cond_actions"],
            block["cond_type"],
            block["cond_value"],
            block["cond_value_size"],
        ) = parse_conditional(block["expr"])
        if "last_block" not in block:
            block["action_list"] = [
                actionD.ActionD(
                    # If this isn't the last block, len(blocks) > 1 so param_skip_all is initialized.
                    expression.ConstantNumeric(param_skip_all),  # lgtm [py/uninitialized-local-variable]
                    expression.ConstantNumeric(0xFF),
                    nmlop.ASSIGN,
                    expression.ConstantNumeric(0),
                    expression.ConstantNumeric(0),
                )
            ]
        else:
            block["action_list"] = []
        for stmt in block["statements"]:
            block["action_list"].extend(stmt.get_action_list())

    # Main problem: action10 can't be skipped by action9, so we're
    # nearly forced to use action7, but action7 can't safely skip action6
    # Solution: use temporary parameter, set to 0 for not skip, !=0 for skip.
    # then skip every block of actions (as large as possible) with either
    # action7 or action9, depending on which of the two works.

    for i, block in enumerate(blocks):
        param = block["param_dst"]
        if i == 0:
            action_list.extend(block["cond_actions"])
        else:
            action_list.extend(
                cond_skip_actions(block["cond_actions"], param_skip_all, (2, r"\7="), 0, 4, cond_list.pos)
            )
            if param is None:
                param = param_skip_all
            else:
                action_list.append(
                    actionD.ActionD(
                        expression.ConstantNumeric(block["param_dst"]),
                        expression.ConstantNumeric(block["param_dst"]),
                        nmlop.AND,
                        expression.ConstantNumeric(param_skip_all),
                    )
                )
        action_list.extend(
            cond_skip_actions(
                block["action_list"],
                param,
                block["cond_type"],
                block["cond_value"],
                block["cond_value_size"],
                cond_list.pos,
            )
        )

    if recursive_cond_blocks == 1:
        free_labels.restore()
    recursive_cond_blocks -= 1
    action6.free_parameters.restore()
    return action_list


def parse_loop_block(loop):
    global recursive_cond_blocks
    recursive_cond_blocks += 1
    if recursive_cond_blocks == 1:
        free_labels.save()

    action6.free_parameters.save()
    begin_label = free_labels.pop_unique(loop.pos)
    action_list = [action10.Action10(begin_label)]

    cond_param, cond_actions, cond_type, cond_value, cond_value_size = parse_conditional(loop.expr)
    block_actions = []
    for stmt in loop.statements:
        block_actions.extend(stmt.get_action_list())

    action_list.extend(cond_actions)
    block_actions.append(UnconditionalSkipAction(9, begin_label))
    action_list.extend(cond_skip_actions(block_actions, cond_param, cond_type, cond_value, cond_value_size, loop.pos))

    if recursive_cond_blocks == 1:
        free_labels.restore()
    recursive_cond_blocks -= 1
    action6.free_parameters.restore()
    return action_list



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import expression, generic, global_constants
from nml.actions import action2, action2layout, action2real, real_sprite
from nml.ast import base_statement, sprite_container


class TemplateDeclaration(base_statement.BaseStatement):
    def __init__(self, name, param_list, sprite_list, pos):
        base_statement.BaseStatement.__init__(self, "template declaration", pos, False, False)
        self.name = name
        self.param_list = param_list
        self.sprite_list = sprite_list

    def pre_process(self):
        # check that all templates that are referred to exist at this point
        # This prevents circular dependencies
        for sprite in self.sprite_list:
            if isinstance(sprite, real_sprite.TemplateUsage):
                if sprite.name.value == self.name.value:
                    raise generic.ScriptError(
                        "Sprite template '{}' includes itself.".format(sprite.name.value), self.pos
                    )
                elif sprite.name.value not in real_sprite.sprite_template_map:
                    raise generic.ScriptError(
                        "Encountered unknown template identifier: " + sprite.name.value, sprite.pos
                    )
        # Register template
        if self.name.value not in real_sprite.sprite_template_map:
            real_sprite.sprite_template_map[self.name.value] = self
        else:
            raise generic.ScriptError(
                "Template named '{}' is already defined, first definition at {}".format(
                    self.name.value, real_sprite.sprite_template_map[self.name.value].pos
                ),
                self.pos,
            )

    def get_labels(self):
        labels = {}
        offset = 0
        for sprite in self.sprite_list:
            sprite_labels, num_sprites = sprite.get_labels()
            for lbl, lbl_offset in sprite_labels.items():
                if lbl in labels:
                    raise generic.ScriptError("Duplicate label encountered; '{}' already exists.".format(lbl), self.pos)
                labels[lbl] = lbl_offset + offset
            offset += num_sprites
        return labels, offset

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Template declaration:", self.name.value)
        generic.print_dbg(indentation + 2, "Parameters:")
        for param in self.param_list:
            param.debug_print(indentation + 4)
        generic.print_dbg(indentation + 2, "Sprites:")
        for sprite in self.sprite_list:
            sprite.debug_print(indentation + 4)

    def get_action_list(self):
        return []

    def __str__(self):
        ret = "template {}({}) {{\n".format(str(self.name), ", ".join([str(param) for param in self.param_list]))
        for sprite in self.sprite_list:
            ret += "\t{}\n".format(sprite)
        ret += "}\n"
        return ret


spriteset_base_class = action2.make_sprite_group_class(True, True, False, cls_is_relocatable=True)


class SpriteSet(spriteset_base_class, sprite_container.SpriteContainer):
    def __init__(self, param_list, sprite_list, pos):
        base_statement.BaseStatement.__init__(self, "spriteset", pos, False, False)
        if not (1 <= len(param_list) <= 5):
            raise generic.ScriptError("Spriteset requires 1 to 5 parameters, encountered " + str(len(param_list)), pos)
        name = param_list[0]
        if not isinstance(name, expression.Identifier):
            raise generic.ScriptError("Spriteset parameter 1 'name' should be an identifier", name.pos)
        sprite_container.SpriteContainer.__init__(self, "spriteset", name)
        self.initialize(name)
        self.zoom_level = 0
        self.bit_depth = 8
        self.image_file = None

        if len(param_list) >= 3:
            if isinstance(param_list[1], expression.Identifier) and param_list[1].value in global_constants.zoom_levels:
                self.zoom_level = global_constants.zoom_levels[param_list[1].value]
            else:
                raise generic.ScriptError(
                    "value for Spriteset-block parameter 2 'zoom level' is not a valid zoom level", param_list[1].pos
                )
            if isinstance(param_list[2], expression.Identifier) and param_list[2].value in global_constants.bit_depths:
                self.bit_depth = global_constants.bit_depths[param_list[2].value]
            else:
                raise generic.ScriptError(
                    "value for Spriteset-block parameter 3 'bit depth' is not a valid bit depth", param_list[2].pos
                )

            if len(param_list) >= 4:
                self.image_file = param_list[3].reduce()
                if not isinstance(self.image_file, expression.StringLiteral):
                    raise generic.ScriptError(
                        "Spriteset-block parameter 4 'file' must be a string literal", self.image_file.pos
                    )
        elif len(param_list) >= 2:
            self.image_file = param_list[1].reduce()
            if not isinstance(self.image_file, expression.StringLiteral):
                raise generic.ScriptError(
                    "Spriteset-block parameter 2 'file' must be a string literal", self.image_file.pos
                )

        if self.bit_depth == 32:
            global_constants.any_32bpp_sprites = global_constants.allow_32bpp

        if len(param_list) >= 5:
            self.mask_file = param_list[4].reduce()
            if not isinstance(self.mask_file, expression.StringLiteral):
                raise generic.ScriptError(
                    "Spriteset-block parameter 5 'mask_file' must be a string literal", self.mask_file.pos
                )
            if not self.bit_depth == 32:
                raise generic.ScriptError("A mask file may only be specified for 32 bpp sprites.", self.mask_file.pos)
        else:
            self.mask_file = None

        self.sprite_list = sprite_list
        self.action1_num = None  # set number in action1
        self.labels = {}  # mapping of real sprite labels to offsets
        self.add_sprite_data(self.sprite_list, self.image_file, pos, self.zoom_level, self.bit_depth, self.mask_file)

    def pre_process(self):
        spriteset_base_class.pre_process(self)
        offset = 0
        for sprite in self.sprite_list:
            sprite_labels, num_sprites = sprite.get_labels()
            for lbl, lbl_offset in sprite_labels.items():
                if lbl in self.labels:
                    raise generic.ScriptError("Duplicate label encountered; '{}' already exists.".format(lbl), self.pos)
                self.labels[lbl] = lbl_offset + offset
            offset += num_sprites

    def collect_references(self):
        return []

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Sprite set:", self.name.value)
        generic.print_dbg(
            indentation + 2, "Source:  ", self.image_file.value if self.image_file is not None else "None"
        )

        generic.print_dbg(indentation + 2, "Sprites:")
        for sprite in self.sprite_list:
            sprite.debug_print(indentation + 4)

    def get_action_list(self):
        # Actions are created when parsing the action2s, not here
        return []

    def __str__(self):
        params = [
            self.name,
            generic.reverse_lookup(global_constants.zoom_levels, self.zoom_level),
            generic.reverse_lookup(global_constants.bit_depths, self.bit_depth),
        ]
        if self.image_file is not None:
            params.append(self.image_file)
        if self.mask_file is not None:
            params.append(self.mask_file)
        ret = "spriteset({}) {{\n".format(", ".join(str(p) for p in params))
        for sprite in self.sprite_list:
            ret += "\t{}\n".format(str(sprite))
        ret += "}\n"
        return ret


spritegroup_base_class = action2.make_sprite_group_class(False, True, False)


class SpriteGroup(spritegroup_base_class):
    def __init__(self, name, spriteview_list, pos=None):
        base_statement.BaseStatement.__init__(self, "spritegroup", pos, False, False)
        self.initialize(name)
        self.spriteview_list = spriteview_list

    def pre_process(self):
        for spriteview in self.spriteview_list:
            spriteview.pre_process()
        spritegroup_base_class.pre_process(self)

    def collect_references(self):
        return []

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Sprite group:", self.name.value)
        for spriteview in self.spriteview_list:
            spriteview.debug_print(indentation + 2)

    def get_action_list(self):
        action_list = []
        if self.prepare_act2_output():
            for feature in sorted(self.feature_set):
                action_list.extend(action2real.get_real_action2s(self, feature))
        return action_list

    def __str__(self):
        ret = "spritegroup {} {{\n".format(self.name)
        for spriteview in self.spriteview_list:
            ret += "\t{}\n".format(spriteview)
        ret += "}\n"
        return ret


class SpriteView:
    def __init__(self, name, spriteset_list, pos):
        self.name = name
        self.spriteset_list = spriteset_list
        self.pos = pos

    def pre_process(self):
        self.spriteset_list = [x.reduce(global_constants.const_list) for x in self.spriteset_list]
        for sg_ref in self.spriteset_list:
            if not (
                isinstance(sg_ref, expression.SpriteGroupRef)
                and action2.resolve_spritegroup(sg_ref.name).is_spriteset()
            ):
                raise generic.ScriptError("Expected a sprite set reference", sg_ref.pos)
            if len(sg_ref.param_list) != 0:
                raise generic.ScriptError(
                    "Spritesets referenced from a spritegroup may not have parameters.", sg_ref.pos
                )

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Sprite view:", self.name.value)
        generic.print_dbg(indentation + 2, "Sprite sets:")
        for spriteset in self.spriteset_list:
            spriteset.debug_print(indentation + 4)

    def __str__(self):
        return "{}: [{}];".format(self.name, ", ".join([str(spriteset) for spriteset in self.spriteset_list]))


spritelayout_base_class = action2.make_sprite_group_class(False, True, False)


class SpriteLayout(spritelayout_base_class):
    def __init__(self, name, param_list, layout_sprite_list, pos=None):
        base_statement.BaseStatement.__init__(self, "spritelayout", pos, False, False)
        self.initialize(name, None, len(param_list))
        self.param_list = param_list
        self.register_map = {}  # Set during action generation for easier referencing
        self.layout_sprite_list = layout_sprite_list

    # Do not reduce expressions here as they may contain variables
    # And the feature is not known yet
    def pre_process(self):
        # Check parameter names
        seen_names = set()
        for param in self.param_list:
            if not isinstance(param, expression.Identifier):
                raise generic.ScriptError("spritelayout parameter names must be identifiers.", param.pos)
            if param.value in seen_names:
                raise generic.ScriptError("Duplicate parameter name '{}' encountered.".format(param.value), param.pos)
            seen_names.add(param.value)
        spritelayout_base_class.pre_process(self)

    def collect_references(self):
        return []

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Sprite layout:", self.name.value)
        generic.print_dbg(indentation + 2, "Parameters:")
        for param in self.param_list:
            param.debug_print(indentation + 4)
        generic.print_dbg(indentation + 2, "Sprites:")
        for layout_sprite in self.layout_sprite_list:
            layout_sprite.debug_print(indentation + 4)

    def __str__(self):
        params = "" if not self.param_list else "({})".format(", ".join(str(x) for x in self.param_list))
        return "spritelayout {}{} {{\n{}\n}}\n".format(
            str(self.name), params, "\n".join(str(x) for x in self.layout_sprite_list)
        )

    def get_action_list(self):
        action_list = []
        if self.prepare_act2_output():
            for feature in sorted(self.feature_set):
                if feature == 0x04:
                    continue
                action_list.extend(action2layout.get_layout_action2s(self, feature))
        return action_list


class LayoutSprite:
    def __init__(self, ls_type, param_list, pos):
        self.type = ls_type
        self.param_list = param_list
        self.pos = pos

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Tile layout sprite of type:", self.type)
        for layout_param in self.param_list:
            layout_param.debug_print(indentation + 2)

    def __str__(self):
        return "\t{} {{\n\t\t{}\n\t}}".format(
            self.type, "\n\t\t".join(str(layout_param) for layout_param in self.param_list)
        )



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import generic

from .base_expression import Expression


class SpecialParameter(Expression):
    """
    Class for handling special grf parameters.
    These can be assigned special, custom methods for reading / writing to them.

    @ivar name: Name of the parameter, for debugging purposes.
    @type name: C{str}

    @ivar info: Information about the parameter.
    @type info: C{dict}

    @ivar write_func: Function that will be called when the parameter is the target of an assignment
                        Arguments:
                            Dictionary with parameter information (self.info)
                            Target expression to assign
                            Position information
                        Return value is a 2-tuple:
                            Left side of the assignment (must be a parameter)
                            Right side of the assignment (can be any expression)
    @type write_func: C{function}

    @ivar read_func: Function that will be called to read out the parameter value
                        Arguments:
                            Dictionary with parameter information (self.info)
                            Position information
                        Return value:
                            Expression that should be evaluated to get the parameter value
    @type read_func: C{function}

    @ivar is_bool: Does read_func return a boolean value?
    @type is_bool: C{bool}
    """

    def __init__(self, name, info, write_func, read_func, is_bool, pos=None):
        Expression.__init__(self, pos)
        self.name = name
        self.info = info
        self.write_func = write_func
        self.read_func = read_func
        self.is_bool = is_bool

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Special parameter '{}'".format(self.name))

    def __str__(self):
        return self.name

    def reduce(self, id_dicts=None, unknown_id_fatal=True):
        return self

    def is_boolean(self):
        return self.is_bool

    def can_assign(self):
        return self.write_func is not None

    def to_assignment(self, expr):
        param, expr = self.write_func(self.info, expr, self.pos)
        param = param.reduce()
        expr = expr.reduce()
        return (param, expr)

    def to_reading(self):
        param = self.read_func(self.info, self.pos)
        param = param.reduce()
        return param

    def supported_by_actionD(self, raise_error):
        return True

    def supported_by_action2(self, raise_error):
        return True
